"""A collection of thin lens equations for system modeling."""

from .mathops import engine as e
from .util import guarantee_array
from .zernike import defocus as _defocus


def object_to_image_dist(efl, object_distance):
    """Compute the image distance from the object distance.

    Parameters
    ----------
    efl : `float`
        focal length of the lens
    object_distance : `float` or `numpy.ndarray`
        distance from the object to the front principal plane of the lens,
        negative for an object to the left of the lens

    Returns
    -------
    `float`
        image distance.  Distance from rear principal plane (assumed to be in
        contact with front principal plane) to image.

    Notes
    -----
    efl and object distance should be in the same units.  Return value will
    be in the same units as the inputs.

    """
    object_distance = guarantee_array(object_distance)
    ret = 1 / efl + 1 / object_distance
    return 1 / ret


def image_to_object_dist(efl, image_distance):
    """Compute the object distance from the image distance.

    Parameters
    ----------
    efl : `float`
        focal length of the lens
    object_distance : `float` or `numpy.ndarray`
        distance from the object to the front principal plane of the lens,
        positive for an object in front of a lens of positive focal length.

    Notes
    -----
    efl and image distance should be in the same units.  Return value will
    be in the same units as the input.

    """
    image_distance = guarantee_array(image_distance)
    ret = 1 / efl - 1 / image_distance
    return 1 / ret


def image_dist_epd_to_na(image_distance, epd):
    """Compute the NA from an image distance and entrance pupil diameter.

    Parameters
    ----------
    image_distance : `float`
        distance from the image to the entrance pupil
    epd : `float`
        diameter of the entrance pupil

    Returns
    -------
    `float`
        numerical aperture.  The NA of the system.

    """
    image_distance = guarantee_array(image_distance)

    rho = epd / 2
    marginal_ray_angle = abs(e.arctan2(rho, image_distance))
    return marginal_ray_angle


def image_dist_epd_to_fno(image_distance, epd):
    """Compute the f/# from an image distance and entrance pupil diameter.

    Parameters
    ----------
    image_distance : `float`
        distance from the image to the entrance pupil
    epd : `float`
        diameter of the entrance pupil

    Returns
    -------
    `float`
        fno.  The working f/# of the system.

    """
    na = image_dist_epd_to_na(image_distance, epd)
    return na_to_fno(na)


def fno_to_na(fno):
    """Convert an fno to an NA.

    Parameters
    ----------
    fno : `float`
        focal ratio

    Returns
    -------
    `float`
        NA.  The NA of the system.

    """
    return 1 / (2 * fno)


def na_to_fno(na):
    """Convert an NA to an f/#.

    Parameters
    ----------
    na : `float`
        numerical aperture

    Returns
    -------
    `float`
        fno.  The f/# of the system.

    """
    return 1 / (2 * e.sin(na))


def object_dist_to_mag(efl, object_dist):
    """Compute the linear magnification from the object distance and focal length.

    Parameters
    ----------
    efl : `float`
        focal length of the lens
    object_dist : `float`
        object distance

    Returns
    -------
    `float`
        linear magnification.  Also known as the lateral magnification

    """
    object_dist = guarantee_array(object_dist)
    return efl / (efl - object_dist)


def mag_to_object_dist(efl, mag):
    """Compute the object distance for a given focal length and magnification.

    Parameters
    ----------
    efl : `float`
        focal length of the lens
    mag : `float`
        signed magnification

    Returns
    -------
    `float`
        object distance

    """
    return efl * ((1/mag) + 1)


def linear_to_long_mag(lateral_mag):
    """Compute the longitudinal (along optical axis) magnification from the lateral mag.

    Parameters
    ----------
    lateral_mag : `float`
        linear magnification, from thin lens formulas

    Returns
    -------
    `float`
        longitudinal magnification

    """
    return lateral_mag**2


def mag_to_fno(mag, infinite_fno, pupil_mag=1):
    """Compute the working f/# from the magnification and infinite f/#.

    Parameters
    ----------
    mag : `float` or `numpy.ndarray`
        linear or lateral magnification
    infinite_fno : `float`
        f/# as defined by EFL/EPD
    pupil_mag : `float`
        pupil magnification

    Returns
    -------
    `float`
        working f/number

    """
    mag = guarantee_array(mag)
    return (1 + abs(mag) / pupil_mag) * infinite_fno


def defocus_to_image_displacement(defocus, fno, wavelength, zernike=False, norm=False):
    """Compute image displacment from wavefront defocus expressed in waves 0-P to.

    Parameters
    ----------
    defocus : `float` or `numpy.ndarray`
        wavefront defocus
    fno : `float`
        f/# of the lens or system
    wavelength : `float`
        wavelength of light, expressed in micron
    zernike : `bool`
        zernike model of defocus (otherwise model is Seidel)
    norm : `bool`
        if zernike model, term is rms normalized

    Returns
    -------
    `float`
        image displacement.  Motion of image in um caused by defocus OPD

    """
    defocus = guarantee_array(defocus)

    # if the defocus is a zernike, make it match Seidel notation for equation validity
    if zernike is True:
        if norm is True:
            defocus = defocus * _defocus.norm  # not using *= on these to avoid side effects with in-place ops
        defocus = defocus * 2
    return 8 * fno**2 * wavelength * defocus


def image_displacement_to_defocus(image_displacement, fno, wavelength, zernike=False, norm=False):
    """Compute the wavefront defocus from image shift, expressed in the same units as the shift.

    Parameters
    ----------
    image_displacement : `float` or ~`numpy.ndarray`
        displacement of the image
    fno : `float`
        f/# of the lens or system
    wavelength : `float`
        wavelength of light, expressed in microns
    zernike : `bool`
        return in Zernike notation
    norm : `bool`
        subset of zernike -- return rms normalized zernike

    Returns
    -------
    `float`
        wavefront defocus

    """
    image_displacement = guarantee_array(image_displacement)
    defocus = image_displacement / (8 * fno ** 2 * wavelength)
    if zernike is True:
        if norm is True:
            return defocus / 2 / _defocus.norm
        else:
            return defocus / 2
    else:
        return defocus


def twolens_efl(efl1, efl2, separation):
    """Use thick lens equations to compute the focal length for two elements separated by some distance.

    Parameters
    ----------
    efl1 : `float`
        EFL of the first lens
    efl2 : `float`
        EFL of the second lens

    separation : `float`
        separation of the two lenses

    Returns
    -------
    `float`
        focal length of the two lens system

    """
    phi1, phi2, t = 1 / efl1, 1 / efl2, separation
    phi_tot = phi1 + phi2 - t * phi1 * phi2
    return 1 / phi_tot


def twolens_bfl(efl1, efl2, separation):
    """Use thick lens equations to compute the back focal length for two elements separated by some distance.

    Parameters
    ----------
    efl1 : `float`
        EFL of the first lens
    efl2 : `float`
        EFL of the second lens

    separation : `float`
        separation of the two lenses.

    Returns
    -------
    `float`
        back focal length of the two lens system.

    """
    phi1 = 1 / efl1
    numerator = 1 - separation * phi1
    denomenator = twolens_efl(efl1, efl2, separation)
    return numerator / denomenator



"""Configuration for this instance of prysm."""
import copy

import numpy as np

from astropy import units as u

from .wavelengths import HeNe

all_ap_unit_types = (u.Unit, u.core.IrreducibleUnit, u.core.CompositeUnit)


def sanitize_unit(unit, wavelength):
    """Sanitize a unit token, either an astropy unit or a string.

    Parameters
    ----------
    unit : `astropy.Unit` or `str`
        unit or string version of unit
    wavelength : `astropy.Unit`
        a wavelength unit generated by mkwvl or equivalent code

    Returns
    -------
    `astropy.Unit`
        an astropy unit

    """
    if not isinstance(unit, all_ap_unit_types):
        if unit.lower() in ('waves', 'wave', 'λ'):
            unit = wavelength
        else:
            unit = getattr(u, unit)
    else:
        unit = unit

    return unit


def format_unit(unit_or_quantity, fmt):
    """(string) format a unit or quantity

    Parameters
    ----------
    unit_or_quantity : `astropy.units.Unit` or `astropy.units.Quantity`
        a unit or quantity
    fmt : `str`, {'latex', 'unicode'}
        a string format

    Returns
    -------
    `str`
        string

    """
    if isinstance(unit_or_quantity, all_ap_unit_types):
        return unit_or_quantity.to_string(fmt)
    elif isinstance(unit_or_quantity, u.quantity.Quantity):
        return unit_or_quantity.unit.to_string(fmt)
    else:
        raise ValueError('must be a Unit or Quantity instance.')


class Labels:
    """Labels holder for data instances."""
    def __init__(self, xy_base, z,
                 xy_additions=['X', 'Y'], xy_addition_side='right',
                 addition_joiner=' ',
                 unit_prefix='[',
                 unit_suffix=']',
                 unit_joiner=' '):
        """Create a new Labels instance

        Parameters
        ----------
        xy_base : `str`
            basic string used to build the X and Y labels
        z : `str`
            z label, stored as self._z to avoid clash with self.z()
        xy_additions : iterable, optional
            text to add to the (x, y) labels
        xy_addition_side : {'left', 'right'. 'l', 'r'}, optional
            side to add the x and y additional text to, left or right
        addition_joiner : `str`, optional
            text used to join the x or y addition
        unit_prefix : `str`, optional
            prefix used to surround the unit text
        unit_suffix : `str`, optional
            suffix used to surround the unit text
        unit_joiner : `str`, optional
            text used to combine the base label and the unit
        """
        self.xy_base, self._z = xy_base, z
        self.xy_additions, self.xy_addition_side = xy_additions, xy_addition_side
        self.addition_joiner = addition_joiner
        self.unit_prefix, self.unit_suffix = unit_prefix, unit_suffix
        self.unit_joiner = unit_joiner

    def _label_factory(self, label, xy_unit, z_unit):
        """Factory method to produce complex labels.

        Parameters
        ----------
        label : `str`, {'x', 'y', 'z'}
            label to produce

        Returns
        -------
        `str`
            completed label

        """
        if label in ('x', 'y'):
            if label == 'x':
                xy_pos = 0
            else:
                xy_pos = 1
            label_basics = [self.xy_base]
            if self.xy_addition_side.lower() in ('left', 'l'):
                label_basics.insert(0, self.xy_additions[xy_pos])
            else:
                label_basics.append(self.xy_additions[xy_pos])

            label_ = self.addition_joiner.join(label_basics)
            unit_str = format_unit(xy_unit, config.unit_format)
        else:
            label_ = self._z
            unit_str = format_unit(z_unit, config.unit_format)

        unit_text = ''
        if config.show_units:

            unit_text = unit_text.join([self.unit_prefix,
                                       unit_str,
                                       self.unit_suffix])
        label_ = self.unit_joiner.join([label_, unit_text])
        return label_

    def x(self, xy_unit, z_unit):
        """X label."""
        return self._label_factory('x', xy_unit, z_unit)

    def y(self, xy_unit, z_unit):
        """Y label."""
        return self._label_factory('y', xy_unit, z_unit)

    def z(self, xy_unit, z_unit):
        """Z label."""
        return self._label_factory('z', xy_unit, z_unit)

    def generic(self, xy_unit, z_unit):
        """Generic label without extra X/Y annotation."""
        base = self.xy_base
        join = self.unit_joiner
        unit = format_unit(xy_unit, config.unit_format)
        prefix = self.unit_prefix
        suffix = self.unit_suffix
        return f'{base}{join}{prefix}{unit}{suffix}'

    def copy(self):
        return copy.deepcopy(self)


rel = u.def_unit(['rel'], format={'latex': 'Rel 1.0', 'unicode': 'Rel 1.0'})

default_phase_units = {'xy': u.mm, 'z': u.nm}
default_interferorgam_units = {'xy': u.pixel, 'z': u.nm}
default_image_units = {'xy': u.mm, 'z': u.adu}
default_mtf_units = {'xy': u.mm ** -1, 'z': rel}
default_ptf_units = {'xy': u.mm ** -1, 'z': u.deg}

xi_eta = ['ξ', 'η']
x_y = ['X', 'Y']
default_pupil_labels = Labels(xy_base='Pupil', z='OPD', xy_additions=xi_eta)
default_interferogram_labels = Labels(xy_base='', z='Height', xy_additions=x_y)
default_convolvable_labels = Labels(xy_base='Image Plane', z='Irradiance', xy_additions=x_y)
default_mtf_labels = Labels(xy_base='Spatial Frequency', z='MTF', xy_additions=x_y)
default_ptf_labels = Labels(xy_base='Spatial Frequency', z='PTF', xy_additions=xi_eta)
default_psd_labels = Labels(xy_base='Spatial Frequency', z='PSD', xy_additions=x_y)


class Config(object):
    """Global configuration of prysm."""
    def __init__(self,
                 precision=64,
                 backend=np,
                 zernike_base=1,
                 Q=2,
                 wavelength=HeNe,
                 phase_cmap='inferno',
                 image_cmap='Greys_r',
                 lw=3,
                 zorder=3,
                 alpha=1,
                 interpolation='lanczos',
                 unit_format='latex_inline',
                 show_units=True,
                 phase_xy_unit=u.mm,
                 phase_z_unit=u.nm,
                 image_xy_unit=u.um,
                 image_z_unit=u.adu,
                 mtf_xy_unit=u.mm ** -1,
                 mtf_z_unit=rel,
                 ptf_xy_unit=u.mm ** -1,
                 ptf_z_unit=u.deg,
                 pupil_labels=default_pupil_labels,
                 interferogram_labels=default_interferogram_labels,
                 convolvable_labels=default_convolvable_labels,
                 mtf_labels=default_mtf_labels,
                 ptf_labels=default_ptf_labels,
                 psd_labels=default_psd_labels):
        """Create a new `Config` object.

        Parameters
        ----------
        precision : `int`
            32 or 64, number of bits of precision
        backend : `str`, {'np'}
            a supported backend.  Current options are only "np" for numpy
        zernike_base : `int`, {0, 1}
            base for zernikes; start at 0 or 1
        Q : `float`
            oversampling parameter for numerical propagations
        phase_cmap : `str`
            colormap used for plotting optical phases
        image_cmap : `str`
            colormap used for plotting greyscale images
        lw : `float`
            linewidth
        zorder : `int`, optional
            zorder used for graphics made with matplotlib
        interpolation : `str`
            interpolation type for 2D plots
        unit_formatter : `str`, optional
            string passed to astropy.units.(unit).to_string
        xylabel_joiner : `str`, optional
            text used to glue together X/Y units and their basic string
        unit_prefix : `str`, optional
            text preceeding the unit's representation, after the joiner
        unit_suffix : `str`, optional
            text following the unit's representation
        unit_joiner : `str`, optional
            text used to glue basic labels and the units together
        show_units : `bool`, optional
            if True, shows units on graphics
        phase_units : `Units`
            default units used for phase-like types
        image_units : `Units`
            default units used for image-like types

        """
        self.initialized = False
        self.precision = precision
        self.backend = backend
        self.zernike_base = zernike_base
        self.chbackend_observers = []
        self.Q = Q
        self.wavelength = wavelength
        self.phase_cmap = phase_cmap
        self.image_cmap = image_cmap
        self.lw = lw
        self.zorder = zorder
        self.alpha = alpha
        self.interpolation = interpolation
        self.unit_format = unit_format
        self.show_units = show_units
        self.phase_xy_unit = phase_xy_unit
        self.phase_z_unit = phase_z_unit
        self.image_xy_unit = image_xy_unit
        self.image_z_unit = image_z_unit
        self.mtf_xy_unit = mtf_xy_unit
        self.mtf_z_unit = mtf_z_unit
        self.ptf_xy_unit = ptf_xy_unit
        self.ptf_z_unit = ptf_z_unit
        self.pupil_labels = pupil_labels
        self.interferogram_labels = interferogram_labels
        self.convolvable_labels = convolvable_labels
        self.mtf_labels = mtf_labels
        self.ptf_labels = ptf_labels
        self.psd_labels = psd_labels
        self.initialized = True

    @property
    def precision(self):
        """Precision used for computations.

        Returns
        -------
        `object` : `numpy.float32` or `numpy.float64`
            precision used

        """
        return self._precision

    @property
    def precision_complex(self):
        """Precision used for complex array computations.

        Returns
        -------
        `object` : `numpy.complex64` or `numpy.complex128`
            precision used for complex arrays

        """
        return self._precision_complex

    @precision.setter
    def precision(self, precision):
        """Adjust precision used by prysm.

        Parameters
        ----------
        precision : `int`, {32, 64}
            what precision to use; either 32 or 64 bits

        Raises
        ------
        ValueError
            if precision is not a valid option

        """
        if precision not in (32, 64):
            raise ValueError('invalid precision.  Precision should be 32 or 64.')

        if precision == 32:
            self._precision = np.float32
            self._precision_complex = np.complex64
        else:
            self._precision = np.float64
            self._precision_complex = np.complex128

    @property
    def backend(self):
        """Backend used.

        Returns
        -------
        `str`
            {'np'} only

        """
        return self._backend

    @backend.setter
    def backend(self, backend):
        """Set the backend used by prysm.

        Parameters
        ----------
        backend : `str`, {'np'}
            backend used for computations

        Raises
        ------
        ValueError
            invalid backend

        """
        if isinstance(backend, str):
            if backend.lower() in ('np', 'numpy'):
                backend = 'numpy'
            elif backend.lower() in ('cp', 'cu', 'cuda'):
                backend = 'cupy'

            exec(f'import {backend}')
            self._backend = eval(backend)
        else:
            self._backend = backend

        if self.initialized:
            for obs in self.chbackend_observers:
                obs(self._backend)

    @property
    def zernike_base(self):
        """Zernike base.

        Returns
        -------
        `int`
            {0, 1}

        """
        return self._zernike_base

    @zernike_base.setter
    def zernike_base(self, base):
        """Zernike base; base-0 or base-1.

        Parameters
        ----------
        base : `int`, {0, 1}
            first index of zernike polynomials

        Raises
        ------
        ValueError
            invalid base given

        """
        if base not in (0, 1):
            raise ValueError('By convention zernike base must be 0 or 1.')

        self._zernike_base = base


config = Config()



"""Defines behavior of convolvable items and a base class to encapsulate that behavior."""
import types

from .mathops import engine as e
from ._richdata import RichData
from .coordinates import resample_2d_complex
from .conf import config
from .fttools import forward_ft_unit, pad2d


class Convolvable(RichData):
    """A base class for convolvable objects to inherit from."""
    _data_attr = 'data'
    _data_type = 'image'

    def __init__(self, x, y, data, has_analytic_ft=False, labels=None, xy_unit=None, z_unit=None):
        """Create a new Convolvable object.

        Parameters
        ----------
        x : `numpy.ndarray`
            1D ndarray defining x data grid
        y : `numpy.ndarray`
            1D ndarray defining y data grid
        data : `numpy.ndarray`
            2D ndarray of data
        has_analytic_ft : `bool`, optional
            Whether this convolvable overrides self.analytic_ft, and has a known
            analytical fourier tansform
        labels : `Labels`
            labels to use.  If None, will use config.convolvable_labels

        """
        xy_unit = 'um'
        z_unit = 'adu'
        super().__init__(x=x, y=y, data=data,
                         xy_unit=xy_unit or config.image_xy_unit,
                         z_unit=z_unit or config.image_z_unit,
                         labels=labels or config.convolvable_labels)
        self.has_analytic_ft = has_analytic_ft

    def __str__(self):
        """Pretty print description."""
        return f'{type(self)} with sample spacing {self.sample_spacing:.3f} and support {self.support:.3f} μm'

    @property
    def support_x(self):
        """Width of the domain in X."""
        return (self.samples_x - 1) * self.sample_spacing

    @property
    def support_y(self):
        """Width of the domain in Y."""
        return (self.samples_y - 1) * self.sample_spacing

    @property
    def support(self):
        """Width of the domain."""
        return max((self.support_x, self.support_y))

    def conv(self, other):
        """Convolves this convolvable with another.

        Parameters
        ----------
        other : `Convolvable`
            A convolvable object

        Returns
        -------
        `Convolvable`
            a convolvable object

        Notes
        -----
        If self and other both have analytic Fourier transforms, no math will be done and the aFTs
        are merged directly.

        If only one of self or other has an analytic Fourier transform, the output grid will be
        defined by the object which does not have an analytic Fourier transform.

        If neither has an analytic transform, the output grid will:
        - span max(self.support, other.support)
        - have sample spacing min(self.sample_spacing, other.sample_spacing)

        This ensures the signal remains Nyquist sampled and (probably) doesn't expand beyond
        the extent of the output window.  The latter condition will be violated when two large
        convolvables are convolved.

        """
        e = ConvolutionEngine(self, other)
        return e.fire()

    def deconv(self, other, balance=1000, reg=None, is_real=True, clip=False, postnormalize=True):
        """Perform the deconvolution of this convolvable object by another.

        Parameters
        ----------
        other : `Convolvable`
            another convolvable object, used as the PSF in a Wiener deconvolution
        balance : `float`, optional
            regularization parameter; passed through to skimage
        reg : `numpy.ndarray`, optional
            regularization operator, passed through to skimage
        is_real : `bool`, optional
            True if self and other are both real
        clip : `bool`, optional
            clips self and other into (0,1)
        postnormalize : `bool`, optional
            normalize the result such that it falls in [0,1]


        Returns
        -------
        `Convolvable`
            a new Convolable object

        Notes
        -----
        See skimage:
        http://scikit-image.org/docs/dev/api/skimage.restoration.html#skimage.restoration.wiener

        """
        from skimage.restoration import wiener

        result = wiener(self.data, other.data, balance=balance, reg=reg, is_real=is_real, clip=clip)
        if postnormalize:
            result += result.min()
            result /= result.max()
        return Convolvable(result, self.x, self.y, False)

    def renorm(self):
        """Renormalize so that the peak is at a value of unity and the minimum value is zero"""
        self.data -= self.data.min()
        self.data /= self.data.max()
        return self

    def msaa(self, factor=2):
        """Multi-Sample anti-aliasing

        Perform anti-aliasing by averaging blocks of (factor, factor) pixels
        into a simple value.

        Parameters
        ----------
        factor : `int`, optional
            factor by which to decimate the data

        Returns
        -------
        `Convolvable`
            self

        """
        from .detector import bindown
        x, y, data = self.x, self.y, self.data
        data = bindown(data, factor, factor, 'avg')
        self.data = data
        self.x = x[::factor]
        self.y = y[::factor]
        return self

    def save(self, path, nbits=8):
        """Write the image to a png, jpg, tiff, etc.

        Parameters
        ----------
        path : `string`
            path to write the image to
        nbits : `int`
            number of bits in the output image

        """
        from imageio import imwrite
        if nbits == 8:
            typ = e.uint8
        elif nbits == 16:
            typ = e.uint16
        else:
            raise ValueError('must use either 8 or 16 bpp.')
        dat = (self.data * 2**nbits - 1).astype(typ)
        imwrite(path, dat)

    @staticmethod
    def from_file(path, scale):
        """Read a monochrome 8 bit per pixel file into a new Image instance.

        Parameters
        ----------
        path : `string`
            path to a file
        scale : `float`
            pixel scale, in microns

        Returns
        -------
        `Convolvable`
            a new image object

        """
        from imageio import imread
        imgarr = imread(path)
        s = imgarr.shape
        extx, exty = (s[1] * scale) / 2, (s[0] * scale) / 2
        ux, uy = e.arange(-extx, extx, scale), e.arange(-exty, exty, scale)
        return Convolvable(data=e.flip(imgarr, axis=0).astype(config.precision),
                           x=ux, y=uy, has_analytic_ft=False)


class ConvolutionEngine:
    """An engine to facilitate fine-grained control over convolutions."""
    def __init__(self, c1, c2=None, spatial_finalization=(abs,), Q=2, pad_method='linear_ramp'):
        self.c1 = c1
        self.c2 = c2
        self.spatial_finalization = spatial_finalization
        self.Q = Q
        self.pad_method = pad_method

        self.spatial_x = None
        self.spatial_y = None
        self.spatial_data = None
        self.kspace_x = None
        self.kspace_y = None
        self.kspace_data = None

        self.nsamples_x = None
        self.nsamples_y = None
        self.sample_spacing = None

    def fire(self):
        """Convolve self.c1 and self.c2 with no fuss."""
        try:
            return self.merge_analytics()
        except ValueError:
            self.compute_kspace_units()
            self.compute_kspace_data()
            self.compute_spatial_units()
            self.ifft()
            self.crop_output()
            self.postprocess_spatial()
            return Convolvable(*self.spatial, has_analytic_ft=False)

    def compute_kspace_data(self):
        """Compute the k-space representation of the convolution of c1 and c2."""
        if self.c1.has_analytic_ft:
            # units came directly from c2, pad and FT c1
            c2_pad = pad2d(self.c2.data, self.Q, mode=self.pad_method)
            c2_ft = e.fft.fftshift(e.fft.fft2(e.fft.ifftshift(c2_pad)))
            xx, yy = e.meshgrid(self.kspace_x, self.kspace_y)
            c1_ft = self.c1.analytic_ft(xx, yy)
        elif self.c2.has_analytic_ft:
            # units came directly from c1, pad and FT c2
            c1_pad = pad2d(self.c1.data, self.Q, mode=self.pad_method)
            c1_ft = e.fft.fftshift(e.fft.fft2(e.fft.ifftshift(c1_pad)))
            xx, yy = e.meshgrid(self.kspace_x, self.kspace_y)
            c2_ft = self.c2.analytic_ft(xx, yy)
        else:
            need_to_interp_c1 = False
            need_to_interp_c2 = False
            # units came from both, need to do some interpolation
            cutoff_c1 = 1 / (2 * self.c1.sample_spacing)
            cutoff_c2 = 1 / (2 * self.c2.sample_spacing)
            cutoff = max(cutoff_c1, cutoff_c2)
            support = max(self.c1.support, self.c2.support)
            if not (self.c1.support == support and cutoff_c1 == cutoff):
                need_to_interp_c1 = True
            if not (self.c2.support == support and cutoff_c2 == cutoff):
                need_to_interp_c2 = True

            def resample_data(self, data, sample_spacing):
                c_freq_x = forward_ft_unit(sample_spacing, data.shape[1])
                c_freq_y = forward_ft_unit(sample_spacing, data.shape[0])
                return resample_2d_complex(data,
                                           (c_freq_x, c_freq_y),
                                           (self.kspace_x, self.kspace_y))

            c1_pad = pad2d(self.c1.data, self.Q, mode=self.pad_method)
            c1_ft = e.fft.fftshift(e.fft.fft2(e.fft.ifftshift(c1_pad)))

            c2_pad = pad2d(self.c2.data, self.Q, mode=self.pad_method)
            c2_ft = e.fft.fftshift(e.fft.fft2(e.fft.ifftshift(c2_pad)))

            if need_to_interp_c1:
                c1_ft = resample_data(self, c1_ft, self.c1.sample_spacing)
            if need_to_interp_c2:
                c2_ft = resample_data(self, c2_ft, self.c2.sample_spacing)

        self.kspace_data = c1_ft * c2_ft
        return self

    def compute_kspace_units(self):
        """Compute the k-space domain of the convolution of c1 and c2."""
        if self.c1.has_analytic_ft:
            support_x, support_y = self.c2.support_x, self.c2.support_y
            sample_spacing = self.c2.sample_spacing
        elif self.c2.has_analytic_ft:
            support_x, support_y = self.c1.support_x, self.c1.support_y
            sample_spacing = self.c1.sample_spacing
        else:
            support_x = max(self.c1.support_x, self.c2.support_x)
            support_y = max(self.c1.support_y, self.c2.support_y)
            sample_spacing = min(self.c1.sample_spacing, self.c2.sample_spacing)

        self.sample_spacing = sample_spacing
        self.nsamples_x = int(e.floor(round(((support_x / sample_spacing) + 1) * self.Q, 6)))
        self.nsamples_y = int(e.floor(round(((support_y / sample_spacing) + 1) * self.Q, 6)))
        self.kspace_x = forward_ft_unit(sample_spacing, self.nsamples_x, True)
        self.kspace_y = forward_ft_unit(sample_spacing, self.nsamples_y, True)
        return self

    def compute_spatial_units(self):
        """Compute the spatial domain units of the convolution of c1 and c2."""
        dx = -1 / (2 * self.kspace_x[0])  # [0] is -fs/2, [-1] is slightly below fs/2 for even-length arrays
        dy = -1 / (2 * self.kspace_y[0])
        ny, nx = self.kspace_data.shape
        support_x, support_y = dx * nx, dy * ny
        self.spatial_x = e.linspace(-support_x/2, support_x/2, nx)
        self.spatial_y = e.linspace(-support_y/2, support_y/2, ny)
        return self

    def ifft(self):
        """Take the iFT to compute the spatial representation of the convolution of c1 and c2."""
        self.spatial_data = e.fft.fftshift(e.fft.ifft2(e.fft.ifftshift(self.kspace_data)))
        return self

    def crop_output(self):
        """Crop the output in the spatial domain to remove the padded area."""
        s = self.kspace_data.shape
        npx_x, npx_y = s[1] / self.Q / 2, s[0] / self.Q / 2
        cx_left, cx_right = int(e.ceil(npx_x)), int(e.floor(npx_x))
        cy_top, cy_bottom = int(e.ceil(npx_y)), int(e.floor(npx_y))
        self.spatial_data = self.spatial_data[cy_top:-cy_bottom, cx_left:-cx_right]
        self.spatial_x = self.spatial_x[cx_left:-cx_right]
        self.spatial_y = self.spatial_y[cy_top:-cy_bottom]
        return self

    def postprocess_spatial(self):
        """Post-process the spatial domain."""
        if self.spatial_finalization is not None:
            for func in self.spatial_finalization:
                self.spatial_data = func(self.spatial_data)

    def merge_analytics(self):
        """Merge c1 and c2 if they both have analytic FTs, else raise.

        Raises
        ------
        ValueError
            c1 or c2 does not have an analytic FT.

        """
        if not (self.c1.has_analytic_ft and self.c2.has_analytic_ft):
            raise ValueError('both convolvables must have analytic FTs')
        else:
            c_out = Convolvable(None, None, None, has_analytic_ft=True)
            c_out.s1 = self.c1.copy()
            c_out.s2 = self.c2.copy()

            def aft(self, x, y):
                part1 = self.s1.analytic_ft(x, y)
                part2 = self.s2.analytic_ft(x, y)
                return part1 * part2

            c_out.analytic_ft = types.MethodType(aft, c_out)
            return c_out

    @property
    def spatial(self):
        """Spatial representation, x, y, data."""
        return self.spatial_x, self.spatial_y, self.spatial_data

    @property
    def kspace(self):
        """k-space representation, fx, fy, data."""
        return self.kspace_x, self.kspace_y, self.kspace_data



"""Sample data for prysm tests and documentation."""
import shutil
from pathlib import Path
from urllib.request import urlopen

baseremote = r'https://github.com/brandondube/prysm/raw/master/sample_files/'
baselocal = Path(__file__)
root = baselocal.parent / '..' / 'prysm-sampledata'
root.mkdir(exist_ok=True)


def fetch_if_not_present(local, remote):
    """Fetch a file from github if it is not present in the local filesystem."""
    if not local.exists():
        with urlopen(remote) as response, open(local, 'wb') as fid:
            shutil.copyfileobj(response, fid)

    return local


class SampleFiles:
    """Sample files for prysm."""
    dat = 'valid_zygo_dat_file.dat'
    mtfvfvf = 'valid_sample_MTFvFvF_Sag.txt'
    mtfvf = 'valid_sample_trioptics_mtf_vs_field.mht'
    mtf = 'valid_sample_trioptics_mtf.mht'

    def __call__(self, dtype_or_filename):
        """Get the path of a sample file."""
        dtype_or_filename = dtype_or_filename.lower()
        if hasattr(self, dtype_or_filename):
            filename = getattr(self, dtype_or_filename)
            local = (root / getattr(self, dtype_or_filename)).absolute()
            remote = baseremote + filename
            return fetch_if_not_present(local, remote)

        else:
            local = root / dtype_or_filename
            remote = baseremote + dtype_or_filename
            return fetch_if_not_present(local, remote)


sample_files = SampleFiles()



"""Objects for image simulation with."""

from scipy.signal import chirp

from .conf import config
from .mathops import engine as e, jinc
from .convolution import Convolvable
from .coordinates import cart_to_polar, polar_to_cart


class Slit(Convolvable):
    """Representation of a slit or pair of slits."""

    def __init__(self, width, orientation='Vertical', sample_spacing=None, samples=0):
        """Create a new Slit instance.

        Parameters
        ----------
        width : `float`
            the width of the slit in microns
        orientation : `string`, {'Horizontal', 'Vertical', 'Crossed', 'Both'}
            the orientation of the slit; Crossed and Both produce the same results
        sample_spacing : `float`
            spacing of samples in the synthetic image
        samples : `int`
            number of samples per dimension in the synthetic image

        Notes
        -----
        Default of 0 samples allows quick creation for convolutions without
        generating the image; use samples > 0 for an actual image.

        """
        w = width / 2

        if samples > 0:
            ext = samples / 2 * sample_spacing
            x = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
            y = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
            arr = e.zeros((samples, samples))
        else:
            arr, x, y = None, None, None

        # paint in the slit
        if orientation.lower() in ('v', 'vert', 'vertical'):
            if samples > 0:
                arr[:, abs(x) < w] = 1
            self.orientation = 'Vertical'
            self.width_x = width
            self.width_y = 0
        elif orientation.lower() in ('h', 'horiz', 'horizontal'):
            if samples > 0:
                arr[abs(y) < w, :] = 1
            self.width_x = 0
            self.width_y = width
            self.orientation = 'Horizontal'
        elif orientation.lower() in ('b', 'both', 'c', 'crossed'):
            if samples > 0:
                arr[abs(y) < w, :] = 1
                arr[:, abs(x) < w] = 1
            self.orientation = 'Crossed'
            self.width_x, self.width_y = width, width

        super().__init__(data=arr, x=x, y=y, has_analytic_ft=True)

    def analytic_ft(self, x, y):
        """Analytic fourier transform of a slit.

        Parameters
        ----------
        x : `numpy.ndarray`
            sample points in x frequency axis
        y : `numpy.ndarray`
            sample points in y frequency axis

        Returns
        -------
        `numpy.ndarray`
            2D numpy array containing the analytic fourier transform

        """
        if self.width_x > 0 and self.width_y > 0:
            return (e.sinc(x * self.width_x) +
                    e.sinc(y * self.width_y)).astype(config.precision)
        elif self.width_x > 0 and self.width_y == 0:
            return e.sinc(x * self.width_x).astype(config.precision)
        else:
            return e.sinc(y * self.width_y).astype(config.precision)


class Pinhole(Convolvable):
    """Representation of a pinhole."""
    def __init__(self, width, sample_spacing=None, samples=0):
        """Create a Pinhole instance.

        Parameters
        ----------
        width : `float`
            the width of the pinhole
        sample_spacing : `float`
            spacing of samples in the synthetic image
        samples : `int`
            number of samples per dimension in the synthetic image

        Notes
        -----
        Default of 0 samples allows quick creation for convolutions without
        generating the image; use samples > 0 for an actual image.

        """
        self.width = width

        # produce coordinate arrays
        if samples > 0:
            ext = samples / 2 * sample_spacing
            x = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
            y = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
            xv, yv = e.meshgrid(x, y)
            w = width / 2
            # paint a circle on a black background
            arr = e.zeros((samples, samples))
            arr[e.sqrt(xv**2 + yv**2) < w] = 1
        else:
            arr, x, y = None, None, None

        super().__init__(data=arr, x=x, y=y, has_analytic_ft=True)

    def analytic_ft(self, x, y):
        """Analytic fourier transform of a slit.

        Parameters
        ----------
        x : `numpy.ndarray`
            sample points in x frequency axis
        y : `numpy.ndarray`
            sample points in y frequency axis

        Returns
        -------
        `numpy.ndarray`
            2D numpy array containing the analytic fourier transform

        """
        # factor of pi corrects for jinc being modulo pi
        # factor of 2 converts radius to diameter
        rho = e.sqrt(x**2 + y**2) * self.width * 2 * e.pi
        return jinc(rho).astype(config.precision)


class SiemensStar(Convolvable):
    """Representation of a Siemen's star object."""
    def __init__(self, spokes, sinusoidal=True, radius=0.9, background='black', sample_spacing=2, samples=256):
        """Produce a Siemen's Star.

        Parameters
        ----------
        spokes : `int`
            number of spokes in the star.
        sinusoidal : `bool`
            if True, generates a sinusoidal Siemen' star, else, generates a bar/block siemen's star
        radius : `float`,
            radius of the star, relative to the array width (default 90%)
        background : 'string', {'black', 'white'}
            background color
        sample_spacing : `float`
            spacing of samples, in microns
        samples : `int`
            number of samples per dimension in the synthetic image

        Raises
        ------
        ValueError
            background other than black or white

        """
        self.spokes = spokes
        self.radius = radius

        # generate a coordinate grid
        x = e.linspace(-1, 1, samples, dtype=config.precision)
        y = e.linspace(-1, 1, samples, dtype=config.precision)
        xx, yy = e.meshgrid(x, y)
        rv, pv = cart_to_polar(xx, yy)
        ext = sample_spacing * (samples / 2)
        ux = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
        uy = e.arange(-ext, ext, sample_spacing, dtype=config.precision)

        # generate the siemen's star as a (rho,phi) polynomial
        arr = e.cos(spokes / 2 * pv)

        if not sinusoidal:  # make binary
            arr[arr < 0] = -1
            arr[arr > 0] = 1

        # scale to (0,1) and clip into a disk
        arr = (arr + 1) / 2
        if background.lower() in ('b', 'black'):
            arr[rv > radius] = 0
        elif background.lower() in ('w', 'white'):
            arr[rv > radius] = 1
        else:
            raise ValueError('invalid background color')

        super().__init__(data=arr, x=ux, y=uy, has_analytic_ft=False)


class TiltedSquare(Convolvable):
    """Represents a tilted square for e.g. slanted-edge MTF calculation."""
    def __init__(self, angle=4, background='white', sample_spacing=2, samples=256, radius=0.3, contrast=0.9):
        """Create a new TitledSquare instance.

        Parameters
        ----------
        angle : `float`
            angle in degrees to tilt w.r.t. the x axis
        background : `string`
            white or black; the square will be the opposite color of the background
        sample_spacing : `float`
            spacing of samples
        samples : `int`
            number of samples
        radius : `float`
            fractional
        contrast : `float`
            contrast, anywhere from 0 to 1

        """
        if background.lower() == 'white':
            arr = e.ones((samples, samples), dtype=config.precision)
            fill_with = 1 - contrast
        else:
            arr = e.zeros((samples, samples), dtype=config.precision)
            fill_with = 1

        ext = samples / 2 * sample_spacing
        radius = radius * ext * 2
        x = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
        y = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
        xx, yy = e.meshgrid(x, y)

        # TODO: convert inline operation to use of rotation matrix
        angle = e.radians(angle)
        xp = xx * e.cos(angle) - yy * e.sin(angle)
        yp = xx * e.sin(angle) + yy * e.cos(angle)
        mask = (abs(xp) < radius) * (abs(yp) < radius)
        arr[mask] = fill_with
        super().__init__(data=arr, x=x, y=y, has_analytic_ft=False)


class SlantedEdge(Convolvable):
    """Representation of a slanted edge."""
    def __init__(self, angle=4, contrast=0.9, crossed=False, sample_spacing=2, samples=256):
        """Create a new TitledSquare instance.

        Parameters
        ----------
        angle : `float`
            angle in degrees to tilt w.r.t. the y axis
        contrast : `float`
            difference between minimum and maximum values in the image
        crossed : `bool`, optional
            whether to make a single edge (crossed=False) or pair of crossed edges (crossed=True)
            aka a "BMW target"
        sample_spacing : `float`
            spacing of samples
        samples : `int`
            number of samples

        """
        diff = (1 - contrast) / 2
        arr = e.full((samples, samples), 1 - diff)
        ext = samples / 2 * sample_spacing
        x = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
        y = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
        xx, yy = e.meshgrid(x, y)

        angle = e.radians(angle)
        xp = xx * e.cos(angle) - yy * e.sin(angle)
        # yp = xx * e.sin(angle) + yy * e.cos(angle)  # do not need this
        mask = xp > 0  # single edge
        if crossed:
            mask = xp > 0  # set of 4 edges
            upperright = mask & e.rot90(mask)
            lowerleft = e.rot90(upperright, 2)
            mask = upperright | lowerleft

        arr[mask] = diff
        self.contrast = contrast
        self.black = diff
        self.white = 1 - diff
        super().__init__(data=arr, x=x, y=y, has_analytic_ft=False)


class Grating(Convolvable):
    """A grating with a given ruling."""
    def __init__(self, period, angle=0, sinusoidal=False, sample_spacing=2, samples=256):
        """Create a new Grating object

        Parameters
        ----------
        period : `float`
            period of the grating in microns
        angle : `float`, optional
            clockwise angle of the grating w.r.t. the X axis, degrees
        sinusoidal : `bool`, optional
            if True, the grating is a sinusoid, else it has square edges
        sample_spacing : `float`, optional
            center-to-center sample spacing in microns
        samples : `int`, optional
            number of samples across the diameter of the array

        """
        self.period = period
        self.sinusoidal = sinusoidal

        ext = samples / 2 * sample_spacing
        x = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
        y = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
        xx, yy = e.meshgrid(x, y)
        if angle != 0:
            rho, phi = cart_to_polar(xx, yy)
            phi += e.radians(angle)
            xx, yy = polar_to_cart(rho, phi)

        data = e.cos(2 * e.pi / period * xx)
        if sinusoidal:
            data += 1
            data /= 2
        else:
            data[data > 0] = 1
            data[data < 0] = 0

        super().__init__(data=data, x=x, y=y, has_analytic_ft=False)


class GratingArray(Convolvable):
    """An array of gratings with given rulings."""
    def __init__(self, periods, angles=None, sinusoidal=False, sample_spacing=2, samples=256):
        # if angles not provided, angles are 0
        if angles is None:
            angles = [0] * len(periods)

        self.periods = periods
        self.angles = angles
        self.sinusoidal = sinusoidal

        # calculate the basic grid things are defined on
        ext = samples / 2 * sample_spacing
        x = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
        y = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
        xx, yy = e.meshgrid(x, y)
        xxx, yyy = xx, yy

        # compute the grid parameters; number of columns, number of samples per column
        squareness = e.sqrt(len(periods))
        ncols = int(e.ceil(squareness))
        samples_per_patch = int(e.floor(samples / ncols))
        low_idx_x = 0
        high_idx_x = samples_per_patch
        low_idx_y = 0
        high_idx_y = samples_per_patch
        curr_row = 0

        out = e.zeros(xx.shape)
        for idx, (period, angle) in enumerate(zip(periods, angles)):
            # if we're off at an off angle, adjust the coordinates
            if angle != 0:
                rho, phi = cart_to_polar(xxx, yyy)
                phi += e.radians(angle)
                xxx, yyy = polar_to_cart(rho, phi)

            # compute the sinusoid
            data = e.cos(2 * e.pi / period * xxx)

            # compute the indices to embed it into the final array;
            # every time the current column advances, advance the X coordinates
            sy = slice(low_idx_y, high_idx_y)
            sx = slice(low_idx_x, high_idx_x)
            out[sy, sx] += data[sy, sx]

            # advance the indices are needed
            if (idx > 0) & ((idx + 1) % ncols == 0):
                offset = samples_per_patch * curr_row
                low_idx_x = 0
                high_idx_x = samples_per_patch
                low_idx_y = samples_per_patch + offset
                high_idx_y = samples_per_patch * 2 + offset
                curr_row += 1
            else:
                low_idx_x += samples_per_patch
                high_idx_x += samples_per_patch

            xxx = xx

        if sinusoidal:
            out += 1
            out /= 2
        else:
            out[out > 0] = 1
            out[out < 0] = 0
        super().__init__(data=out, x=x, y=y, has_analytic_ft=False)


class Chirp(Convolvable):
    """A frequency chirp."""
    def __init__(self, p0, p1, angle=0, method='linear', binary=True, sample_spacing=2, samples=256, aspect=4):
        """Create a new Chirp instance.

        Parameters
        ----------
        p0 : `float`
            first period, units of microns
        p1 : `float`
            second period, units of microns
        angle : `float`
            clockwise angle between the X axis and the chirp, units of degrees
        method : `str`, optional, {'linear', 'quadratic', 'logarithmic', 'hyperbolic'}
            type of chirp, passed directly to scipy.signal.chirp
        binary : `bool`, optional
            if True, the chirp is a square bar target, not a sinusoidal target.
        sample_spacing : `float`, optional
            center-to-center spacing of samples in the array
        samples : `float`, optional
            number of samples
        aspect : `int`, optional
            aspect ratio, y will have (aspect) times fewer samples than x

        """
        p0 *= 2
        p1 *= 2
        ext = samples / 2 * sample_spacing
        x = e.arange(0, 2 * ext, sample_spacing, dtype=config.precision)
        y = e.arange(0, 2 * ext / aspect, sample_spacing, dtype=config.precision)  # 8:1 aspect ratio
        xx, yy = e.meshgrid(x, y)

        if angle != 0:
            rho, phi = cart_to_polar(xx, yy)
            phi += e.radians(angle)
            xx, yy = polar_to_cart(rho, phi)

        sig = chirp(xx, f0=1 / p0, f1=1 / p1, t1=x[-1], method=method)
        if binary:
            sig[sig < 0] = 0
            sig[sig > 0] = 1
        else:
            sig = (sig + 1) / 2

        super().__init__(x=x, y=y, data=sig, has_analytic_ft=False)



"""High performance / recursive jacobi polynomial calculation."""
from .mathops import engine as e


def weight(alpha, beta, x):
    """The weight function of the jacobi polynomials for a given alpha, beta value."""
    one_minus_x = 1 - x
    return (one_minus_x ** alpha) * (one_minus_x ** beta)


def a(n, alpha, beta, x):
    """The leading term of the recurrence relation from Wikipedia, * P_n^(a,b)."""
    term1 = 2 * n
    term2 = n + alpha + beta
    term3 = 2 * n + alpha + beta - 2
    return term1 * term2 * term3


def b(n, alpha, beta, x):
    """The second term of the recurrence relation from Wikipedia, * P_n-1^(a,b)."""
    term1 = 2 * n + alpha + beta - 1
    iterm1 = 2 * n + alpha + beta
    iterm2 = 2 * n + alpha + beta - 2
    iterm3 = alpha ** 2 - beta ** 2
    temp_product = iterm1 * iterm2 * x + iterm3
    return term1 * temp_product


def c(n, alpha, beta, x):
    """The third term of the recurrence relation from Wikipedia, * P_n-2^(a,b)."""
    term1 = 2 * (n + alpha - 1)
    term2 = (n + beta - 1)
    term3 = (2 * n + alpha + beta)
    return term1 * term2 * term3


def jacobi(n, alpha, beta, x, Pnm1=None, Pnm2=None):
    """Jacobi polynomial of order n with weight parameters alpha and beta.

    Notes
    -----
    This function is faster than scipy.special.jacobi when Pnm1 and Pnm2 are
    supplied and is stable to high order.  Performance benefit ranges from 2-5x.

    Parameters
    ----------
    n : `int`
        polynomial order
    alpha : `float`
        first weight parameter
    beta : `float`
        second weight parameter
    x : `numpy.ndarray`
        x coordinates to evaluate at
    Pnm1 : `numpy.ndarray`, optional
        The n-1th order jacobi polynomial, evaluated at the given points
    Pnm2 : `numpy.ndarray`, optional
        The n-2th order jacobi polynomial, evaluated at the given points

    Returns
    -------
    `numpy.ndarray`
        jacobi polynomial evaluated at the given points

    """
    if n == 0:
        return e.ones_like(x)
    elif n == 1:
        term1 = alpha + 1
        term2 = alpha + beta + 2
        term3 = (x - 1) / 2
        return term1 + term2 * term3
    elif n > 1:
        if Pnm1 is None:
            Pnm1 = jacobi(n-1, alpha, beta, x)
        if Pnm2 is None:
            Pnm2 = jacobi(n-2, alpha, beta, x)

        a_ = a(n, alpha, beta, x)
        b_ = b(n, alpha, beta, x)
        c_ = c(n, alpha, beta, x)
        term1 = b_ * Pnm1
        term2 = c_ * Pnm2
        tmp = term1 - term2
        return tmp / a_


def jacobi_sequence(n_max, alpha, beta, x):
    """Jacobi polynomials of order 0..n_max with weight parameters alpha and beta.

    Parameters
    ----------
    n_max : `int`
        maximum polynomial order
    alpha : `float`
        first weight parameter
    beta : `float`
        second weight parameter
    x : `numpy.ndarray`
        x coordinates to evaluate at

    Returns
    -------
    `numpy.ndarray`
        array of shape (n_max+1, len(x))
    """
    out = e.empty((n_max + 1, len(x)))
    out[0, :] = jacobi(0, alpha, beta, x)
    if n_max > 0:
        out[1, :] = jacobi(1, alpha, beta, x)

    if n_max > 1:
        for i in range(2, n_max + 1):
            Pnm1 = out[i - 1, :]
            Pnm2 = out[i - 2, :]
            out[i, :] = jacobi(i, alpha, beta, x, Pnm1=Pnm1, Pnm2=Pnm2)

    return out



"""Degredations in the image chain."""

from .conf import config
from .mathops import engine as e
from .coordinates import cart_to_polar, polar_to_cart
from .convolution import Convolvable


class Smear(Convolvable):
    """Smear (motion blur)."""
    def __init__(self, width, angle=0):
        """Create a new Smear model.

        Parameters
        ----------
        width : `float`
            width of the blur in microns
        angle : `float`
            clockwise angle of the blur with respect to the x axis in degrees.

        """
        super().__init__(None, None, None, True)
        self.width = width
        self.angle = angle

    def analytic_ft(self, x, y):
        """Analytic FT of the smear.

        Parameters
        ----------
        x : `numpy.ndarray`
            x Cartesian spatial frequency, cy/um
        y : `numpy.ndarray`
            y Cartesian spatial frequency, cy/um

        Returns
        -------
        `numpy.ndarray`
            analytical FT of the smear.

        """
        if self.angle != 0:
            rho, phi = cart_to_polar(x, y)
            phi += e.radians(self.angle)
            x, y = polar_to_cart(rho, phi)

        return e.sinc(x * self.width)


class Jitter(Convolvable):
    """Jitter (high frequency motion)."""
    def __init__(self, scale, sample_spacing=None, samples=None):
        """Create a new Jitter instance.

        Parameters
        ----------
        scale : `float`
            scale of the jitter, units of microns
        sample_spacing : `float`, optional
            center-to-center sample spacing, units of microns
        samples : `int`, optional
            number of samples in X and Y

        """
        self.scale = scale
        if samples is not None:
            ext = (samples - 1) * sample_spacing / 2
            x = e.arange(-ext, ext, sample_spacing, dtype=config.precision)
            y = e.arange(-ext, ext, sample_spacing, dtype=config.precision)

            coef = 1 / (scale * e.sqrt(2 * e.pi))
            xx, yy = e.meshgrid(x, y)
            rho, _ = cart_to_polar(xx, yy)
            kernel = rho ** 2 / (2 * scale ** 2)
            z = coef * e.exp(-kernel)
        else:
            x, y, z = None, None, None

        super().__init__(data=z, x=x, y=y, has_analytic_ft=True)

    def analytic_ft(self, x, y):
        """Analytic FT of jitter.

        Parameters
        ----------
        x : `numpy.ndarray`
            x Cartesian spatial frequency, units of cy/um
        y : `numpy.ndarray`
            y Cartesian spatial frequency, units of cy/um

        Returns
        -------
        `numpy.ndarray`
            value of analytic FT

        """
        rho, _ = cart_to_polar(x, y)
        kernel = e.pi * self.scale / 2 * rho
        return e.exp(-2 * kernel**2)



"""Plotting-related functions."""
from .conf import config


def share_fig_ax(fig=None, ax=None, numax=1, sharex=False, sharey=False):
    """Reurns the given figure and/or axis if given one.  If they are None, creates a new fig/ax.

    Parameters
    ----------
    fig : `matplotlib.figure.Figure`, optional
        figure
    ax : `matplotlib.axes.Axis`
        axis or array of axes
    numax : `int`
        number of axes in the desired figure, 1 for most plots, 3 for plot_fourier_chain
    sharex : `bool`, optional
        whether to share the x axis
    sharey : `bool`, optional
        whether to share the y axis

    Returns
    -------
    `matplotlib.figure.Figure`
        A figure object
    `matplotlib.axes.Axis`
        An axis object

    """
    from matplotlib import pyplot as plt

    if fig is None and ax is None:
        fig, ax = plt.subplots(nrows=1, ncols=numax, sharex=sharex, sharey=sharey)
    elif ax is None:
        ax = fig.gca()

    return fig, ax


def add_psd_model(psd, fig=None, ax=None, invert_x=False,
                  lw=None, ls='--', color='k', alpha=1, zorder=None,
                  psd_fcn=None, **psd_fcn_kwargs):
    """add a PSD model to

    Parameters
    ----------
    psd : `prysm.interferogram.PSD`
        a PSD object
    fig : `matplotlib.figure.Figure`
        Figure containing the plot
    ax : `matplotlib.axes.Axis`
        Axis containing the plot
    invert_x : `bool`, optional
        if True, plot with x axis of spatial period
    lw : `float`, optional
        line width
    color : `str`, optional
        something matplotlib understands as a color
    alpha : `float`, optional
        alpha (transparency) parameter for matplotlib
    zorder : `int`, optional
        z order (height in the stack)
    psd_fcn : `callable`, optional
        a callable function.  If None, inferred between ab_psd and abc_psd based on if c is in psd_fcn_kwargs
    **psd_fcn_kwargs
        keyword arguments arguments passed to psd_fcn after the spatial frequency variable

    Returns
    -------
    fig : `matplotlib.figure.Figure`
        Figure containing the plot
    ax : `matplotlib.axes.Axis`
        Axis containing the plot

    """
    if lw is None:
        lw = config.lw * 1.5

    if zorder is None:
        zorder = config.zorder

    u = psd.slices().x[0]
    if invert_x:
        u2 = u.copy()
        u2[0] = u2[1] / 4
        u = 1 / u2

    fig, ax = share_fig_ax(fig, ax)

    if psd_fcn is None:
        from .interferogram import abc_psd, ab_psd
        if 'c' in psd_fcn_kwargs:
            psd_fcn = abc_psd
        else:
            psd_fcn = ab_psd

    line = psd_fcn(u2, **psd_fcn_kwargs)

    ax.plot(u, line, lw=lw, color=color, alpha=alpha, zorder=zorder)

    return fig, ax



"""Detector-related simulations."""
from collections import deque

from .conf import config
from .mathops import engine as e
from .convolution import Convolvable
from .util import is_odd


class Detector(object):
    """Model of a image sensor."""

    def __init__(self, pitch_x=None, pitch_y=None, pixel='rectangle',
                 resolution=(1024, 1024), nbits=16, framebuffer=10):
        """Create a new Detector object.

        Parameters
        ----------
        pixel_size : `float`
            size of pixels, in um
        resolution : `iterable`
            (x,y) resolution in pixels
        nbits : `int`
            number of bits to digitize to
        framebuffer : `int`
            number of frames of data to store

        """
        if isinstance(pixel, str):
            pixel = pixel.lower()
            if pixel == 'rectangle' and pitch_x is None and pitch_y is None:
                raise ValueError('must provide at least x pitch for rectangular pixels.')

            if pixel == 'rectangle':
                pixel = PixelAperture(pitch_x, pitch_y)
                self.rectangular_100pct_fillfactor_pix = True
        else:
            self.rectangular_100pct_fillfactor_pix = False

        self.pixel = pixel

        if pitch_y is None:
            pitch_y = pitch_x

        if not hasattr(resolution, '__iter__'):
            resolution = (resolution, resolution)

        self.pitch_x = pitch_x
        self.pitch_y = pitch_y
        self.resolution = resolution
        self.bit_depth = nbits
        self.captures = deque(maxlen=framebuffer)

    def capture(self, convolvable):
        """Sample a convolvable, mimics capturing a photo of an oversampled representation of an image.

        Parameters
        ----------
        convolvable : `prysm.Convolvable`
            a convolvable object

        Returns
        -------
        `prysm.convolvable`
            a new convolvable object, as it would be sampled by the detector

        Raises
        ------
        ValueError
            if the convolvable would have to become supersampled by the detector;
            this would lead to an inaccurate result and is not supported

        """
        ss = convolvable.sample_spacing
        pitch_x_err = abs(self.pitch_x % ss) / ss
        pitch_y_err = abs(self.pitch_y % ss) / ss

        ptol = 0.01  # 1%
        if (self.rectangular_100pct_fillfactor_pix
           and (pitch_x_err < ptol)
           and (pitch_y_err < ptol)):
            ux, uy, data = bindown_with_units(self.pitch_x,
                                              self.pitch_y,
                                              convolvable.sample_spacing,
                                              convolvable.data)
            c_out = Convolvable(data=data, x=ux, y=uy, has_analytic_ft=False)
        else:
            from skimage.transform import resize
            c_out = self.pixel.conv(convolvable)
            ss = c_out.sample_spacing
            py, px = c_out.shape
            oy = int(e.floor(py * (ss / self.pitch_y)))
            ox = int(e.floor(px * (ss / self.pitch_x)))

            # resize combines decimation and interpolation and is an effective resampler
            out_data = resize(c_out.data, (oy, ox), mode='reflect', anti_aliasing=False, clip=False, order=3)

            oext_x = (ox - 1) * self.pitch_x / 2
            oext_y = (oy - 1) * self.pitch_y / 2
            out_x = e.arange(ox) * self.pitch_x - oext_x
            out_y = e.arange(oy) * self.pitch_y - oext_y
            c_out = Convolvable(data=out_data, x=out_x, y=out_y)

        self.captures.append(c_out)
        return c_out

    def save_image(self, path, which='last'):
        """Save an image captured by the detector.

        Parameters
        ----------
        path : `string`
            path to save the image to

        which : `string` or `int`
            if string, "first" or "last", otherwise index into the capture buffer of the camera.

        Raises
        ------
        ValueError
            bad target frame to save; should always be the a valid int < buffer_depth

        """
        if which.lower() == 'last':
            self.captures[-1].save(path, self.bit_depth)
        elif type(which) is int:
            self.captures[which].save(path, self.bit_depth)
        else:
            raise ValueError('invalid "which" provided')

    def show_image(self, which='last', fig=None, ax=None):
        """Show an image captured by the detector.

        Parameters
        ----------
        which : `string` or `int`
            if string, "first" or "last", otherwise index into the capture buffer of the camera
        fig : `matplotlib.figure.Figure`, optional
            Figure containing the plot
        ax : `matplotlib.axes.Axis`, optional
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        """
        if which.lower() == 'last':
            which = -1

        fig, ax = self.captures[which].plot2d(fig=fig, ax=ax)
        return fig, ax

    @property
    def pitch(self):
        """1D pixel pitch - minimum of x/y pitches."""
        return min(self.pitch_x, self.pitch_y)

    @pitch.setter
    def pitch(self, pitch_x, pitch_y=None):
        """Set the pixel pitch.

        Parameters
        ----------
        pitch_x : `float`
            x axis pixel pitch
        pitch_y : `float`, optional
            y axis pixel pitch, copies x pitch if not given.

        """
        pitch_y = pitch_x or pitch_y
        self.pitch_x = pitch_x
        self.pitch_y = pitch_y

    @property
    def fill_factor_x(self):
        """Fill factor in the X axis."""
        return self.pixel.width_x / self.pitch_x

    @property
    def fill_factor_y(self):
        """Fill factor in the Y axis."""
        return self.pixel.width_y / self.pitch_y

    @property
    def fill_factor(self):
        """1D fill factor -- minimum of x/y fill factors."""
        return min(self.fill_factor_x, self.fill_factor_y)

    @property
    def fs(self):
        """Sampling frequency in cy/mm."""  # NQOA
        return 1 / self.pitch * 1e3

    @property
    def nyquist(self):
        """Nyquist frequency in cy/mm."""
        return self.fs / 2

    @property
    def last(self):
        """Last frame captured."""
        return self.captures[-1]


class OLPF(Convolvable):
    """Optical Low Pass Filter."""

    def __init__(self, width_x, width_y=None, sample_spacing=0, samples_x=None, samples_y=None):
        """Create a new OLPF object.

        Parameters
        ----------
        width_x : `float`
            blur width in the x direction, microns
        width_y : `float`
            blur width in the y direction, microns
        sample_spacing : `float`, optional
            center to center spacing of samples
        samples_x : `int`, optional
            number of samples along x axis
        samples_y : `int`, optional
            number of samples along y axis; duplicates x if None

        """
        # compute relevant spacings
        if width_y is None:
            width_y = width_x
        if samples_y is None:
            samples_y = samples_x

        self.width_x = width_x
        self.width_y = width_y

        if samples_x is None:  # do no math
            data, ux, uy = None, None, None
        else:
            space_x = width_x / 2
            space_y = width_y / 2
            shift_x = int(space_x // sample_spacing)
            shift_y = int(space_y // sample_spacing)
            center_x = samples_x // 2
            center_y = samples_y // 2

            data = e.zeros((samples_x, samples_y))

            data[center_y - shift_y, center_x - shift_x] = 1
            data[center_y - shift_y, center_x + shift_x] = 1
            data[center_y + shift_y, center_x - shift_x] = 1
            data[center_y + shift_y, center_x + shift_x] = 1
            ux = e.linspace(-space_x, space_x, samples_x)
            uy = e.linspace(-space_y, space_y, samples_y)

        super().__init__(data=data, x=ux, y=uy, has_analytic_ft=True)

    def analytic_ft(self, x, y):
        """Analytic fourier transform of a pixel aperture.

        Parameters
        ----------
        x : `numpy.ndarray`
            sample points in x axis
        y : `numpy.ndarray`
            sample points in y axis

        Returns
        -------
        `numpy.ndarray`
            2D numpy array containing the analytic fourier transform

        """
        return (e.cos(2 * self.width_x * x) *
                e.cos(2 * self.width_y * y)).astype(config.precision)


class PixelAperture(Convolvable):
    """The aperture of a rectangular pixel."""
    def __init__(self, width_x, width_y=None, sample_spacing=0, samples_x=None, samples_y=None):
        """Create a new `PixelAperture` object.

        Parameters
        ----------
        width_x : `float`
            width of the aperture in the x dimension, in microns.
        width_y : `float`, optional
            siez of the aperture in the y dimension, in microns
        sample_spacing : `float`, optional
            spacing of samples, in microns
        samples_x : `int`, optional
            number of samples in the x dimension
        samples_y : `int`, optional
            number of samples in the y dimension

        """
        if width_y is None:
            width_y = width_x
        if samples_y is None:
            samples_y = samples_x

        self.width_x = width_x
        self.width_y = width_y

        if samples_x is None:  # do no math
            data, ux, uy = None, None, None
        else:  # build PixelAperture model
            center_x = samples_x // 2
            center_y = samples_y // 2
            half_width = width_x / 2
            half_height = width_y / 2
            steps_x = int(half_width // sample_spacing)
            steps_y = int(half_height // sample_spacing)

            data = e.zeros((samples_x, samples_y))
            data[center_y - steps_y:center_y + steps_y,
                 center_x - steps_x:center_x + steps_x] = 1
            extx, exty = samples_x // 2 * sample_spacing, samples_y // 2 * sample_spacing
            ux, uy = e.linspace(-extx, extx, samples_x), e.linspace(-exty, exty, samples_y)
        super().__init__(data=data, x=ux, y=uy, has_analytic_ft=True)

    def analytic_ft(self, x, y):
        """Analytic fourier transform of a pixel aperture.

        Parameters
        ----------
        x : `numpy.ndarray`
            sample points in x axis
        y : `numpy.ndarray`
            sample points in y axis

        Returns
        -------
        `numpy.ndarray`
            2D numpy array containing the analytic fourier transform

        """
        return pixelaperture_analytic_otf(self.width_x, self.width_y, x, y)


def pixelaperture_analytic_otf(width_x, width_y, freq_x, freq_y):
    """Analytic MTF of a rectangular pixel aperture.

    Parameters
    ----------
    width_x : `float`
        x diameter of the pixel, in microns
    width_y : `float`
        y diameter of the pixel, in microns
    freq_x : `numpy.ndarray`
        x spatial frequency, in cycles per micron
    freq_y : `numpy.ndarray`
        y spatial frequency, in cycles per micron

    Returns
    -------
    `numpy.ndarray`
        MTF of the pixel aperture

    """
    return e.sinc(freq_x * width_x) * e.sinc(freq_y * width_y)


def bindown(array, nsamples_x, nsamples_y=None, mode='avg'):
    """Bin (resample) an array.

    Parameters
    ----------
    array : `numpy.ndarray`
        array of values
    nsamples_x : `int`
        number of samples in x axis to bin by
    nsamples_y : `int`
        number of samples in y axis to bin by.  If None, duplicates value from nsamples_x
    mode : `str`, {'avg', 'sum'}
        sum or avg, how to adjust the output signal

    Returns
    -------
    `numpy.ndarray`
        ndarray binned by given number of samples

    Notes
    -----
    Array should be 2D.  TODO: patch to allow 3D data.

    If the size of `array` is not evenly divisible by the number of samples,
    the algorithm will trim around the border of the array.  If the trim
    length is odd, one extra sample will be lost on the left side as opposed
    to the right side.

    Raises
    ------
    ValueError
        invalid mode

    """
    if nsamples_y is None:
        nsamples_y = nsamples_x

    if nsamples_x == 1 and nsamples_y == 1:
        return array

    # determine amount we need to trim the array
    samples_x, samples_y = array.shape
    total_samples_x = samples_x // nsamples_x
    total_samples_y = samples_y // nsamples_y
    final_idx_x = total_samples_x * nsamples_x
    final_idx_y = total_samples_y * nsamples_y

    residual_x = int(samples_x - final_idx_x)
    residual_y = int(samples_y - final_idx_y)

    # if the amount to trim is symmetric, trim symmetrically.
    if not is_odd(residual_x) and not is_odd(residual_y):
        samples_to_trim_x = residual_x // 2
        samples_to_trim_y = residual_y // 2
        trimmed_data = array[samples_to_trim_x:final_idx_x + samples_to_trim_x,
                             samples_to_trim_y:final_idx_y + samples_to_trim_y]
    # if not, trim more on the left.
    else:
        samples_tmp_x = (samples_x - final_idx_x) // 2
        samples_tmp_y = (samples_y - final_idx_y) // 2
        samples_top = int(e.floor(samples_tmp_y))
        samples_bottom = int(e.ceil(samples_tmp_y))
        samples_left = int(e.ceil(samples_tmp_x))
        samples_right = int(e.floor(samples_tmp_x))
        trimmed_data = array[samples_left:final_idx_x + samples_right,
                             samples_bottom:final_idx_y + samples_top]

    intermediate_view = trimmed_data.reshape(total_samples_x, nsamples_x,
                                             total_samples_y, nsamples_y)

    if mode.lower() in ('avg', 'average', 'mean'):
        output_data = intermediate_view.mean(axis=(1, 3))
    elif mode.lower() == 'sum':
        output_data = intermediate_view.sum(axis=(1, 3))
    else:
        raise ValueError('mode must be average of sum.')

    # trim as needed to make even number of samples.
    # TODO: allow work with images that are of odd dimensions
    px_x, px_y = output_data.shape
    trim_x, trim_y = 0, 0
    if is_odd(px_x):
        trim_x = 1
    if is_odd(px_y):
        trim_y = 1

    return output_data[:px_x - trim_x, :px_y - trim_y]


def bindown_with_units(px_x, px_y, source_spacing, source_data):
    """Perform bindown, returning unit axes and data.

    Parameters
    ----------
    px_x : `float`
        pixel pitch in the x direction, microns
    px_y : `float`
        pixel pitch in the y direction, microns
    source_spacing : `float`
        pixel pitch in the source data, microns
    source_data : `numpy.ndarray`
        ndarray of regularly spaced data

    Returns
    -------
    ux : `numpy.ndarray`
        1D array of sample coordinates in the x direction
    uy : `numpy.ndarray`
        1D array of sample coordinates in the y direction
    data : `numpy.ndarray`
        binned-down data

    """
    # we assume the pixels are bigger than the samples in the source
    spp_x = px_x / source_spacing
    spp_y = px_y / source_spacing
    if min(spp_x, spp_y) < 1:
        raise ValueError('Pixels smaller than samples, bindown not possible.')
    else:
        spp_x, spp_y = int(e.ceil(spp_x)), int(e.ceil(spp_y))

    data = bindown(source_data, spp_x, spp_y, 'avg')
    s = data.shape
    extx, exty = s[0] * px_x // 2, s[1] * px_y // 2
    ux, uy = e.arange(-extx, extx, px_x), e.arange(-exty, exty, px_y)
    return ux, uy, data



"""A submodule which imports and exports math functions from different libraries.

The intend is to make the backend for prysm interoperable, allowing users to
utilize more high performance engines if they have them installed, or fall
back to more widely available options in the case that they do not.
"""
import numpy as np

from scipy.special import j1, j0  # NOQA

from prysm.conf import config

# numba funcs
try:
    from numba import jit, vectorize
    numba_installed = True
except ImportError:
    # if Numba is not installed, create the jit decorator and have it return the
    # original function.
    def jit(signature_or_function=None, locals={}, target='cpu', cache=False, **options):
        """Passthrough duplicate of numba jit.

        Parameters
        ----------
        signature_or_function : None, optional
            a function or signature to compile
        locals : dict, optional
            local variables for the compiled code
        target : str, optional
            architecture to compile for
        cache : bool, optional
            whether to cache compilation to disk
        **options
            various options

        Returns
        -------
        callable
            a callable

        """
        if signature_or_function is None:
            def _jit(function):
                return function
            return _jit
        else:
            return signature_or_function

    vectorize = jit
    numba_installed = False

# cuda
try:
    import cupy as cp
    from cupy import fuse
    assert cp
    cuda_compatible = True
except ImportError:
    cuda_compatible = False

    def fuse(*args, **kwargs):
        if 'func' in kwargs:
            return kwargs['func']
        else:
            def wrapper(function):
                return function

            return wrapper


def jinc(r):
    """Jinc.
    Parameters
    ----------
    r : `number`
        radial distance
    Returns
    -------
    `float`
        the value of j1(x)/x for x != 0, 0.5 at 0
    """
    if r < 1e-8 and r > -1e-8:  # value of jinc for x < 1/2 machine precision  is 0.5
        return 0.5
    else:
        return j1(r) / r


if numba_installed is True:
    # one day split numba jit and numpy jit
    jinc = np.vectorize(jinc)
else:
    jinc = np.vectorize(jinc)


def sign(x):
    """Sign of a number.  Note only works for single values, not arrays."""
    return -1 if x < 0 else 1


def kronecker(i, j):
    """Kronecker delta function, 1 if i = j, otherwise 0."""
    return 1 if i == j else 0


def gamma(n, m):
    """Gamma function."""
    if n == 1 and m == 2:
        return 3 / 8
    elif n == 1 and m > 2:
        mm1 = m - 1
        numerator = 2 * mm1 + 1
        denominator = 2 * (mm1 - 1)
        coef = numerator / denominator
        return coef * gamma(1, mm1)
    else:
        nm1 = n - 1
        num = (nm1 + 1) * (2 * m + 2 * nm1 - 1)
        den = (m + nm1 - 2) * (2 * nm1 + 1)
        coef = num / den
        return coef * gamma(nm1, m)


class MathEngine:
    """An engine allowing an interchangeable backend for mathematical functions."""
    def __init__(self, source=np):
        """Create a new math engine.

        Parameters
        ----------
        source : `module`
            a python module.

        """
        self.source = source

    def __getattr__(self, key):
        """Get attribute.

        Parameters
        ----------
        key : `str` attribute name

        """
        try:
            return getattr(self.source, key)
        except AttributeError:
            # function not found, fall back to numpy
            # this will actually work nicely for numpy 1.16+
            # due to the __array_function__ and __array_ufunc__ interfaces
            # that were implemented
            return getattr(self.source, key)  # this can raise, but we don't *need* to catch

    def change_backend(self, backend):
        """Function to run when changing the backend."""
        self.source = backend


engine = MathEngine()
config.chbackend_observers.append(engine.change_backend)



"""Basic class holding data, used to recycle code."""
import copy
import warnings
from numbers import Number
from collections.abc import Iterable

from scipy import interpolate

from .conf import config, sanitize_unit
from .mathops import engine as e
from .wavelengths import mkwvl
from .coordinates import uniform_cart_to_polar, polar_to_cart
from .plotting import share_fig_ax


def fix_interp_pair(x, y):
    if y is None:
        y = 0

    if x is None:
        x = 0

    if isinstance(x, Iterable) and not isinstance(y, Iterable):
        y = [y] * len(x)
    elif isinstance(y, Iterable) and not isinstance(x, Iterable):
        x = [x] * len(y)

    return x, y


class RichData:
    """Abstract base class holding some data properties."""
    _data_attr = 'data'
    _data_type = 'image'
    _default_twosided = True
    _slice_xscale = 'linear'
    _slice_yscale = 'linear'

    def __init__(self, x, y, data, labels, xy_unit=None, z_unit=None, wavelength=None):
        """Initialize a new BasicData instance.

        Parameters
        ----------
        x : `numpy.ndarray`
            x unit axis
        y : `numpy.ndarray`
            y unit axis
        data : `numpy.ndarray`
            data
        labels : `Labels`
            labels instance, can be shared
        xyunit : `astropy.unit` or `str`, optional
            astropy unit or string which satisfies hasattr(astropy.units, xyunit)
        zunit : `astropy.unit` or `str`, optional
             astropy unit or string which satisfies hasattr(astropy.units, xyunit)
        wavelength : `astropy.unit` or `float`
            astropy unit or quantity or float with implicit units of microns

        Returns
        -------
        RichData
            the instance

        """
        if wavelength is None:
            wavelength = config.wavelength

        self.x, self.y = x, y
        setattr(self, self._data_attr, data)
        self.labels = labels
        self.wavelength = mkwvl(wavelength)
        self.xy_unit = sanitize_unit(xy_unit, self.wavelength)
        self.z_unit = sanitize_unit(z_unit, self.wavelength)
        self.interpf_x, self.interpf_y, self.interpf_2d = None, None, None

    @property
    def shape(self):
        """Proxy to phase or data shape."""
        try:
            return getattr(self, self._data_attr).shape
        except AttributeError:
            return (0, 0)

    @property
    def size(self):
        """Proxy to phase or data size."""
        try:
            return getattr(self, self._data_attr).size
        except AttributeError:
            return 0

    @property
    def samples_x(self):
        """Number of samples in the x dimension."""
        return self.shape[1]

    @property
    def samples_y(self):
        """Number of samples in the y dimension."""
        return self.shape[0]

    @property
    def sample_spacing(self):
        """center-to-center sample spacing."""
        try:
            return self.x[1] - self.x[0]
        except TypeError:
            return e.nan

    @property
    def center_x(self):
        """Center "pixel" in x."""
        return self.samples_x // 2

    @property
    def center_y(self):
        """Center "pixel" in y."""
        return self.samples_y // 2

    def copy(self):
        """Return a (deep) copy of this instance."""
        return copy.deepcopy(self)

    def change_xy_unit(self, to, inplace=True):
        """Change the x/y unit to a new one, scaling the data in the process.

        Parameters
        ----------
        to : `astropy.unit` or `str`
            if not an astropy unit, a string that is a valid attribute of astropy.units.
        inplace : `bool`, optional
            if True, returns self.  Otherwise returns the modified data.

        Returns
        -------
        `RichData`
            self, if inplace=True
        `numpy.ndarray`, `numpy.ndarray`
            x, y from self, if inplace=False

        """
        unit = sanitize_unit(to, self.wavelength)
        coef = self.xy_unit.to(unit)
        x, y = self.x * coef, self.y * coef
        if not inplace:
            return x, y
        else:
            self.x, self.y = x, y
            self.xy_unit = unit
            return self

    def change_z_unit(self, to, inplace=True):
        """Change the z unit to a new one, scaling the data in the process.

        Parameters
        ----------
        to : `astropy.unit` or `str`
            if not an astropy unit, a string that is a valid attribute of astropy.units.
        inplace : `bool`, optional
            if True, returns self.  Otherwise returns the modified data.

        Returns
        -------
        `RichData`
            self, if inplace=True
        `numpy.ndarray`
            data from self, if inplace=False

        """
        unit = sanitize_unit(to, self.wavelength)
        coef = self.z_unit.to(unit)
        modified_data = getattr(self, self._data_attr) * coef
        if not inplace:
            return modified_data
        else:
            setattr(self, self._data_attr, modified_data)
            self.z_unit = unit
            return self

    def slices(self, twosided=None):
        """Create a `Slices` instance from this instance.

        Parameters
        ----------
        twosided : `bool`, optional
            if None, copied from self._default_twosided

        Returns
        -------
        `Slices`
            a Slices object

        """
        if twosided is None:
            twosided = self._default_twosided
        return Slices(getattr(self, self._data_attr), x=self.x, y=self.y,
                      twosided=twosided, x_unit=self.xy_unit, z_unit=self.z_unit, labels=self.labels,
                      xscale=self._slice_xscale, yscale=self._slice_yscale)

    def _make_interp_function_2d(self):
        """Generate a 2D interpolation function for this instance, used in sampling with exact_xy.

        Returns
        -------
        `scipy.interpolate.RegularGridInterpolator`
            interpolator instance.

        """
        if self.interpf_2d is None:
            self.interpf_2d = interpolate.RegularGridInterpolator((self.y, self.x), getattr(self, self._data_attr))

        return self.interpf_2d

    def _make_interp_function_xy1d(self):
        """Generate two interpolation functions for the xy slices.

        Returns
        -------
        self.interpf_x : `scipy.interpolate.interp1d`
            x interpolator
        self.interpf_y : `scipy.interpolate.interp1d`
            y interpolator

        """
        if self.interpf_x is None or self.interpf_y is None:
            ux, x = self.slices().x
            uy, y = self.slices().y

            self.interpf_x = interpolate.interp1d(ux, x)
            self.interpf_y = interpolate.interp1d(uy, y)

        return self.interpf_x, self.interpf_y

    def exact_polar(self, rho, phi=None):
        """Retrieve data at the specified radial coordinates pairs.

        Parameters
        ----------
        r : iterable
            radial coordinate(s) to sample
        phi : iterable
            azimuthal coordinate(s) to sample

        Returns
        -------
        `numpy.ndarray`
            data at the given points

        """
        self._make_interp_function_2d()

        rho, phi = fix_interp_pair(rho, phi)
        x, y = polar_to_cart(rho, phi)
        return self.interpf_2d((x, y), method='linear')

    def exact_xy(self, x, y=None):
        """Retrieve data at the specified X-Y frequency pairs.

        Parameters
        ----------
        x : iterable
            X coordinate(s) to retrieve
        y : iterable
            Y coordinate(s) to retrieve

        Returns
        -------
        `numpy.ndarray`
            data at the given points

        """
        self._make_interp_function_2d()

        x, y = fix_interp_pair(x, y)
        return self.interpf_2d((y, x), method='linear')

    def exact_x(self, x):
        """Return data at an exact x coordinate along the y=0 axis.

        Parameters
        ----------
        x : `number` or `numpy.ndarray`
            x coordinate(s) to return

        Returns
        -------
        `numpy.ndarray`
            ndarray of values

        """
        self._make_interp_function_xy1d()
        return self.interpf_x(x)

    def exact_y(self, y):
        """Return data at an exact y coordinate along the x=0 axis.

        Parameters
        ----------
        y : `number` or `numpy.ndarray`
            y coordinate(s) to return

        Returns
        -------
        `numpy.ndarray`
            ndarray of values

        """
        self._make_interp_function_xy1d()
        return self.interpf_y(y)

    def plot2d(self, xlim=None, ylim=None, clim=None, cmap=None,
               log=False, power=1, interpolation=None,
               show_colorbar=True, show_axlabels=True,
               fig=None, ax=None):
        """Plot the data in 2D.

        Parameters
        ----------
        xlim : `float` or iterable, optional
            x axis limits.  If not iterable, symmetric version of the single value
        ylim : `float` or iterable, optional
            y axis limits.  If None and xlim is not None, copied from xlim.
            If not iterable, symmetric version of the single value.
        clim : iterable, optional
            clim passed directly to matplotlib.
            If None, looked up on self._default_clim.
        cmap : `str`, optional
            colormap to use, passed directly to matplotlib if not None.
            If None, looks up the default cmap for self._data_type on config
        log : `bool`, optional
            if True, plot on a log color scale
        power : `float`, optional
            if not 1, plot on a power stretched color scale
        interpolation : `str`, optional
            interpolation method to use, passed directly to matplotlib
        show_colorbar : `bool`, optional
            if True, draws the colorbar
        show_axlabels : `bool`, optional
            if True, draws the axis labels
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        """
        from matplotlib.colors import PowerNorm, LogNorm
        fig, ax = share_fig_ax(fig, ax)

        # sanitize some inputs
        if cmap is None:
            cmap = getattr(config, f'{self._data_type}_cmap')

        if interpolation is None:
            interpolation = config.interpolation

        if xlim is not None and not isinstance(xlim, Iterable):
            xlim = (-xlim, xlim)

        if ylim is None and xlim is not None:
            ylim = xlim
        elif ylim is not None and not isinstance(ylim, Iterable):
            ylim = (-ylim, ylim)

        if clim is not None and not isinstance(clim, Iterable):
            clim = (-clim, clim)

        norm = None
        if log:
            norm = LogNorm()
        elif power != 1:
            norm = PowerNorm(power)

        im = ax.imshow(getattr(self, self._data_attr),
                       extent=[self.x[0], self.x[-1], self.y[0], self.y[-1]],
                       cmap=cmap,
                       clim=clim,
                       norm=norm,
                       origin='lower',
                       interpolation=interpolation)

        if show_colorbar:
            fig.colorbar(im, label=self.labels.z(self.xy_unit, self.z_unit), ax=ax, fraction=0.046)

        xlab, ylab = None, None
        if show_axlabels:
            xlab = self.labels.x(self.xy_unit, self.z_unit)
            ylab = self.labels.y(self.xy_unit, self.z_unit)
        ax.set(xlabel=xlab, xlim=xlim, ylabel=ylab, ylim=ylim)

        return fig, ax

    @property
    def slice_x(self):
        warnings.warn('.slice_x is deprecated and will be removed in prysm v0.18, please use .slices().x')
        return self.slices().x

    @property
    def slice_y(self):
        warnings.warn('.slice_y is deprecated and will be removed in prysm v0.18, please use .slices().y')
        return self.slices().y


class Slices:
    """Slices of data."""
    def __init__(self, data, x, y, x_unit, z_unit, labels, xscale, yscale, twosided=True):
        self._source = data
        self._source_polar = None
        self._r = None
        self._p = None
        self._x = x
        self._y = y
        self.x_unit, self.z_unit = x_unit, z_unit
        self.labels = labels
        self.xscale, self.yscale = xscale, yscale
        self.center_y, self.center_x = (int(e.ceil(s / 2)) for s in data.shape)
        self.twosided = twosided

    def check_polar_calculated(self):
        """Ensure that the polar representation of the source data has been calculated."""
        if self._source_polar is None:
            rho, phi, polar = uniform_cart_to_polar(self._x, self._y, self._source)
            self._r, self._p = rho, phi
            self._source_polar = polar

    @property
    def x(self):
        """Slice through the Y=0 axis of the data, i.e. along the X axis.

        Returns
        -------
        x : `numpy.ndarray`
            coordinates
        slice : `numpy.ndarray`
            values of the data array at these coordinates

        """
        if self.twosided:
            return self._x, self._source[self.center_y, :]
        else:
            return self._x[self.center_x:], self._source[self.center_y, self.center_x:]

    @property
    def y(self):
        """Slice through the X=0 axis of the data, i.e., along the Y axis.

        Returns
        -------
        y : `numpy.ndarray`
            coordinates
        slice : `numpy.ndarray`
            values of the data array at these coordinates

        """
        if self.twosided:
            return self._y, self._source[:, self.center_x]
        else:
            return self._y[self.center_y:], self._source[self.center_y:, self.center_x]

    @property
    def azavg(self):
        """Azimuthal average of the data.

        Returns
        -------
        rho : `numpy.ndarray`
            coordinates
        slice : `numpy.ndarray`
            values of the data array at these coordinates

        """
        self.check_polar_calculated()
        return self._r, e.nanmean(self._source_polar, axis=0)

    @property
    def azmedian(self):
        """Azimuthal median of the data.

        Returns
        -------
        rho : `numpy.ndarray`
            coordinates
        slice : `numpy.ndarray`
            values of the data array at these coordinates

        """
        self.check_polar_calculated()
        return self._r, e.nanmedian(self._source_polar, axis=0)

    @property
    def azmin(self):
        """Azimuthal minimum of the data.

        Returns
        -------
        rho : `numpy.ndarray`
            coordinates
        slice : `numpy.ndarray`
            values of the data array at these coordinates

        """
        self.check_polar_calculated()
        return self._r, e.nanmin(self._source_polar, axis=0)

    @property
    def azmax(self):
        """Azimuthal maximum of the data.

        Returns
        -------
        rho : `numpy.ndarray`
            coordinates
        slice : `numpy.ndarray`
            values of the data array at these coordinates

        """
        self.check_polar_calculated()
        return self._r, e.nanmax(self._source_polar, axis=0)

    @property
    def azpv(self):
        """Azimuthal PV of the data.

        Returns
        -------
        rho : `numpy.ndarray`
            coordinates
        slice : `numpy.ndarray`
            values of the data array at these coordinates

        """
        r, mx = self.azmax
        r, mn = self.azmin
        return r, mx - mn

    @property
    def azvar(self):
        """Azimuthal variance of the data.

        Returns
        -------
        rho : `numpy.ndarray`
            coordinates
        slice : `numpy.ndarray`
            values of the data array at these coordinates

        """
        self.check_polar_calculated()
        return self._r, e.nanvar(self._source_polar, axis=0)

    @property
    def azstd(self):
        """Azimuthal standard deviation of the data.

        Returns
        -------
        rho : `numpy.ndarray`
            coordinates
        slice : `numpy.ndarray`
            values of the data array at these coordinates

        """
        self.check_polar_calculated()
        return self._r, e.nanstd(self._source_polar, axis=0)

    def plot(self, slices, lw=None, alpha=None, zorder=None, invert_x=False,
             xlim=(None, None), xscale=None,
             ylim=(None, None), yscale=None,
             show_legend=True, show_axlabels=True,
             fig=None, ax=None):
        """Plot slice(s)

        Parameters
        ----------
        slices : `str` or `Iterable`
            if a string, plots a single slice.  Else, plots several slices.
        lw : `float` or `Iterable`, optional
            line width to use for the slice(s).
            If a single value, used for all slice(s).
            If iterable, used pairwise with the slices
        alpha : `float` or `Iterable`, optional
            alpha (transparency) to use for the slice(s).
            If a single value, used for all slice(s).
            If iterable, used pairwise with the slices
        zorder : `int` or `Iterable`, optional
            zorder (stack height) to use for the slice(s).
            If a single value, used for all slice(s).
            If iterable, used pairwise with the slices
        invert_x : `bool`, optional
            if True, flip x (i.e., Freq => Period or vice-versa)
        xlim : `tuple`, optional
            x axis limits
        xscale : `str`, {'linear', 'log'}, optional
            scale used for the x axis
        ylim : `tuple`, optional
            y axis limits
        yscale : `str`, {'linear', 'log'}, optional
            scale used for the y axis
        show_legend : `bool`, optional
            if True, show the legend
        show_axlabels : `bool`, optional
            if True, show the axis labels
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        """
        def safely_invert_x(x, v):
            # these values are unsafe for fp32.  Maybe a bit pressimistic, but that's life
            zeros = abs(x) < 1e-9
            x, v = x.copy(), v.copy()
            x[zeros] = e.nan
            v[zeros] = e.nan
            x = 1 / x
            return x, v

        # error check everything
        if alpha is None:
            alpha = config.alpha

        if lw is None:
            lw = config.lw

        if zorder is None:
            zorder = config.zorder

        if isinstance(slices, str):
            slices = [slices]

        if isinstance(alpha, Number):
            alpha = [alpha] * len(slices)

        if isinstance(lw, Number):
            lw = [lw] * len(slices)

        if isinstance(zorder, int):
            zorder = [zorder] * len(slices)

        if not hasattr(xlim, '__iter__') and self.twosided:
            xlim = (-xlim, xlim)

        fig, ax = share_fig_ax(fig, ax)

        for slice_, alpha, lw, zorder in zip(slices, alpha, lw, zorder):
            u, v = getattr(self, slice_)
            if invert_x:
                u, v = safely_invert_x(u, v)

            ax.plot(u, v, label=slice_, lw=lw, alpha=alpha, zorder=zorder)

        if show_legend:
            ax.legend(title='Slice')

        # the x label has some special text manipulation

        if invert_x:
            xlabel = self.labels.generic(self.x_unit ** -1, self.z_unit)
            # ax.invert_xaxis()
            if 'Period' in xlabel:
                xlabel = xlabel.replace('Period', 'Frequency')
            elif 'Frequency' in xlabel:
                xlabel = xlabel.replace('Frequency', 'Period')
        else:
            # slightly unclean code duplication here
            xlabel = self.labels.generic(self.x_unit, self.z_unit)

        ylabel = self.labels.z(self.x_unit, self.z_unit)

        if not show_axlabels:
            xlabel, ylabel = '', ''

        # z looks wrong here, but z from 2D is y in 1D.
        ax.set(xscale=xscale or self.xscale, xlim=xlim, xlabel=xlabel,
               yscale=yscale or self.yscale, ylim=ylim, ylabel=ylabel)
        if invert_x:
            ax.invert_xaxis()

        return fig, ax



"""Utility functions."""
from operator import itemgetter

from .mathops import engine as e


def is_odd(int):
    """Determine if an interger is odd using binary operations.

    Parameters
    ----------
    int : `int`
        an integer

    Returns
    -------
    `bool`
        true if odd, False if even

    """
    return int & 0x1


def is_power_of_2(value):
    """Check if a value is a power of 2 using binary operations.

    Parameters
    ----------
    value : `number`
        value to check

    Returns
    -------
    `bool`
        true if the value is a power of two, False if the value is no

    Notes
    -----
    c++ inspired implementation, see SO:
    https://stackoverflow.com/questions/29480680/finding-if-a-number-is-a-power-of-2-using-recursion

    """
    if value is 1:
        return False
    else:
        return bool(value and not value & (value - 1))


def fold_array(array, axis=1):
    """Fold an array in half over the given axis and averages.

    Parameters
    ----------
    array : `numpy.ndarray`
        ndarray
    axis : `int`, optional
        axis to fold over

    Returns
    -------
    `numpy.ndarray`
        folded array

    """
    xs, ys = array.shape
    if axis is 1:
        xh = xs // 2
        left_chunk = array[:, :xh]
        right_chunk = array[:, xh:]
        folded_array = e.concatenate((right_chunk[:, :, e.newaxis],
                                     e.flip(e.flip(left_chunk, axis=1),
                                            axis=0)[:, :, e.newaxis]),
                                     axis=2)
    else:
        yh = ys // 2
        top_chunk = array[:yh, :]
        bottom_chunk = array[yh:, :]
        folded_array = e.concatenate((bottom_chunk[:, :, e.newaxis],
                                     e.flip(e.flip(top_chunk, axis=1),
                                            axis=0)[:, :, e.newaxis]),
                                     axis=2)
    return folded_array.mean(axis=2)


def mean(array):
    """Return the mean value of the valid elements of an array.

    Parameters
    ----------
    array : `numpy.ndarray`
        array of values

    Returns
    -------
    `float`
        mean value

    """
    non_nan = e.isfinite(array)
    return array[non_nan].mean()


def pv(array):
    """Return the PV value of the valid elements of an array.

    Parameters
    ----------
    array : `numpy.ndarray`
        array of values

    Returns
    -------
    `float`
        PV of the array
    """
    non_nan = e.isfinite(array)
    return array[non_nan].max() - array[non_nan].min()


def rms(array):
    """Return the RMS value of the valid elements of an array.

    Parameters
    ----------
    array : `numpy.ndarray`
        array of values

    Returns
    -------
    `float`
        RMS of the array

    """
    non_nan = e.isfinite(array)
    return e.sqrt((array[non_nan] ** 2).mean())


def Sa(array):
    """Return the Ra value for the valid elements of an array.

    Parameters
    ----------
    array: `numpy.ndarray`
        array of values

    Returns
    -------
    `float`
        Ra of the array
    """
    non_nan = e.isfinite(array)
    ary = array[non_nan]
    mean = ary.mean()
    return abs(ary - mean).sum() / ary.size


def std(array):
    """Return the standard deviation of the valid elements of an array.

    Parameters
    ----------
    array: `numpy.ndarray`
        array of values

    Returns
    -------
    `float`
        std of the array

    """
    non_nan = e.isfinite(array)
    ary = array[non_nan]
    return ary.std()


def guarantee_array(variable):
    """Guarantee that a varaible is a numpy ndarray and supports -, *, +, and other operators.

    Parameters
    ----------
    variable : `number` or `numpy.ndarray`
        variable to coalesce

    Returns
    -------
    `object`
        an object that  supports * / and other operations with ndarrays

    Raises
    ------
    ValueError
        non-numeric type

    """
    if type(variable) in [float, e.ndarray, e.int32, e.int64, e.float32, e.float64, e.complex64, e.complex128]:
        return variable
    elif type(variable) is int:
        return float(variable)
    elif type(variable) is list:
        return e.asarray(variable)
    else:
        raise ValueError(f'variable is of invalid type {type(variable)}')


def ecdf(x):
    """Compute the empirical cumulative distribution function of a dataset.

    Parameters
    ----------
    x : `iterable`
        Data

    Returns
    -------
    xs : `numpy.ndarray`
        sorted data
    ys : `numpy.ndarray`
        cumulative distribution function of the data

    """
    xs = e.sort(x)
    ys = e.arange(1, len(xs) + 1) / float(len(xs))
    return xs, ys


def sort_xy(x, y):
    """Sorts a pair of x and y iterables, returning arrays in order of ascending x.

    Parameters
    ----------
    x : `iterable`
        a list, numpy ndarray, or other iterable to sort by
    y : `iterable`
        a list, numpy ndarray, or other iterable that is y=f(x)

    Returns
    -------
    sorted_x : iterable
        an iterable containing the sorted x elements
    sorted_y : iterable
        an iterable containing the sorted y elements

    """
    # zip x and y, sort by the 0th element (x) of each tuple in zip()
    _ = sorted(zip(x, y), key=itemgetter(0))
    sorted_x, sorted_y = zip(*_)
    return sorted_x, sorted_y



"""Zernike functions."""
from collections import defaultdict
from functools import partial

from retry import retry

from .conf import config
from .mathops import engine as e, jit, vectorize, fuse, kronecker, sign
from .pupil import Pupil
from .coordinates import make_rho_phi_grid, cart_to_polar, gridcache
from .util import rms, sort_xy, is_odd
from .plotting import share_fig_ax
from .jacobi import jacobi

# See JCW - http://wp.optics.arizona.edu/jcwyant/wp-content/uploads/sites/13/2016/08/ZernikePolynomialsForTheWeb.pdf


def piston(rho, phi):
    """Zernike Piston."""
    return e.ones(rho.shape)


def tip(rho, phi):
    """Zernike Tilt-Y."""
    return rho * e.cos(phi)


def tilt(rho, phi):
    """Zernike Tilt-X."""
    return rho * e.sin(phi)


def defocus(rho, phi):
    """Zernike defocus."""
    return 2 * rho**2 - 1


def primary_astigmatism_00(rho, phi):
    """Zernike primary astigmatism 0°."""
    return rho**2 * e.cos(2 * phi)


def primary_astigmatism_45(rho, phi):
    """Zernike primary astigmatism 45°."""
    return rho**2 * e.sin(2 * phi)


def primary_coma_y(rho, phi):
    """Zernike primary coma Y."""
    return (3 * rho**3 - 2 * rho) * e.cos(phi)


def primary_coma_x(rho, phi):
    """Zernike primary coma X."""
    return (3 * rho**3 - 2 * rho) * e.sin(phi)


def primary_spherical(rho, phi):
    """Zernike primary Spherical."""
    return 6 * rho**4 - 6 * rho**2 + 1


def primary_trefoil_y(rho, phi):
    """Zernike primary trefoil Y."""
    return rho**3 * e.cos(3 * phi)


def primary_trefoil_x(rho, phi):
    """Zernike primary trefoil X."""
    return rho**3 * e.sin(3 * phi)


def secondary_astigmatism_00(rho, phi):
    """Zernike secondary astigmatism 0°."""
    return (4 * rho**4 - 3 * rho**2) * e.cos(2 * phi)


def secondary_astigmatism_45(rho, phi):
    """Zernike secondary astigmatism 45°."""
    return (4 * rho**4 - 3 * rho**2) * e.sin(2 * phi)


def secondary_coma_y(rho, phi):
    """Zernike secondary coma Y."""
    return (10 * rho**5 - 12 * rho**3 + 3 * rho) * e.cos(phi)


def secondary_coma_x(rho, phi):
    """Zernike secondary coma X."""
    return (10 * rho**5 - 12 * rho**3 + 3 * rho) * e.sin(phi)


def secondary_spherical(rho, phi):
    """Zernike secondary spherical."""
    return 20 * rho**6 + - 30 * rho**4 + 12 * rho**2 - 1


def primary_tetrafoil_y(rho, phi):
    """Zernike primary tetrafoil Y."""
    return rho**4 * e.cos(4 * phi)


def primary_tetrafoil_x(rho, phi):
    """Zernike primary tetrafoil X."""
    return rho**4 * e.sin(4 * phi)


def secondary_trefoil_y(rho, phi):
    """Zernike secondary trefoil Y."""
    return (5 * rho**5 - 4 * rho**3) * e.cos(3 * phi)


def secondary_trefoil_x(rho, phi):
    """Zernike secondary trefoil X."""
    return (5 * rho**5 - 4 * rho**3) * e.sin(3 * phi)


def tertiary_astigmatism_00(rho, phi):
    """Zernike tertiary astigmatism 0°."""
    return (15 * rho**6 - 20 * rho**4 + 6 * rho**2) * e.cos(2 * phi)


def tertiary_astigmatism_45(rho, phi):
    """Zernike tertiary astigmatism 45°."""
    return (15 * rho**6 - 20 * rho**4 + 6 * rho**2) * e.sin(2 * phi)


def tertiary_coma_y(rho, phi):
    """Zernike tertiary coma Y."""
    return (35 * rho**7 - 60 * rho**5 + 30 * rho**3 - 4 * rho) * e.cos(phi)


def tertiary_coma_x(rho, phi):
    """Zernike tertiary coma X."""
    return (35 * rho**7 - 60 * rho**5 + 30 * rho**3 - 4 * rho) * e.sin(phi)


def tertiary_spherical(rho, phi):
    """Zernike tertiary spherical."""
    return 70 * rho**8 - 140 * rho**6 + 90 * rho**4 - 20 * rho**2 + 1


def primary_pentafoil_y(rho, phi):
    """Zernike primary pentafoil Y."""
    return rho**5 * e.cos(5 * phi)


def primary_pentafoil_x(rho, phi):
    """Zernike primary pentafoil X."""
    return rho**5 * e.sin(5 * phi)


def secondary_tetrafoil_y(rho, phi):
    """Zernike secondary tetrafoil Y."""
    return (6 * rho**6 - 5 * rho**4) * e.cos(4 * phi)


def secondary_tetrafoil_x(rho, phi):
    """Zernike secondary tetrafoil X."""
    return (6 * rho**6 - 5 * rho**4) * e.sin(4 * phi)


def tertiary_trefoil_y(rho, phi):
    """Zernike tertiary trefoil Y."""
    return (21 * rho**7 - 30 * rho**5 + 10 * rho**3) * e.cos(3 * phi)


def tertiary_trefoil_x(rho, phi):
    """Zernike tertiary trefoil X."""
    return (21 * rho**7 - 30 * rho**5 + 10 * rho**3) * e.cos(3 * phi)


def quaternary_astigmatism_00(rho, phi):
    """Zernike quaternary astigmatism 0°."""
    return (21 * rho**6 - 30 * rho**4 + 10 * rho**2) * e.cos(2 * phi)


def quaternary_astigmatism_45(rho, phi):
    """Zernike quaternary astigmatism 45°."""
    return (21 * rho**6 - 30 * rho**4 + 10 * rho**2) * e.sin(2 * phi)


def quaternary_coma_y(rho, phi):
    """Zernike quaternary coma Y."""
    return (126 * rho**9 - 280 * rho**7 + 210 * rho**5 - 60 * rho**3 + 5 * rho)\
        * e.cos(phi)


def quaternary_coma_x(rho, phi):
    """Zernike quaternary coma X."""
    return (126 * rho**9 - 280 * rho**7 + 210 * rho**5 - 60 * rho**3 + 5 * rho)\
        * e.sin(phi)


def quaternary_spherical(rho, phi):
    """Zernike quaternary spherical."""
    return 252 * rho**10 \
        - 630 * rho**8 \
        + 560 * rho**6 \
        - 210 * rho**4 \
        + 30 * rho**2 \
        - 1


def primary_hexafoil_y(rho, phi):
    """Zernike primary hexafoil Y."""
    return rho**6 * e.cos(6 * phi)


def primary_hexafoil_x(rho, phi):
    """Zernike primary hexafoil X."""
    return rho**6 * e.sin(6 * phi)


def secondary_pentafoil_y(rho, phi):
    """Zernike secondary pentafoil Y."""
    return (7 * rho**7 - 6 * rho**5) * e.cos(5 * phi)


def secondary_pentafoil_x(rho, phi):
    """Zernike secondary pentafoil X."""
    return (7 * rho**7 - 6 * rho**5) * e.sin(5 * phi)


def tertiary_tetrafoil_y(rho, phi):
    """Zernike tertiary tetrafoil Y."""
    return (28 * rho**8 - 42 * rho**6 + 15 * rho**4) * e.cos(4 * phi)


def tertiary_tetrafoil_x(rho, phi):
    """Zernike tertiary tetrafoil X."""
    return (28 * rho**8 - 42 * rho**6 + 15 * rho**4) * e.sin(4 * phi)


def quaternary_trefoil_y(rho, phi):
    """Zernike quaternary trefoil Y."""
    return (84 * rho**9 - 168 * rho**7 + 105 * rho**5 - 20 * rho**3) * e.cos(3 * phi)


def quaternary_trefoil_x(rho, phi):
    """Zernike quaternary trefoil X."""
    return (84 * rho**9 - 168 * rho**7 + 105 * rho**5 - 20 * rho**3) * e.sin(3 * phi)


def quinternary_astigmatism_00(rho, phi):
    """Zernike quinternary astigmatism 0°."""
    return (210 * rho**10 - 504 * rho**8 + 420 * rho**6 - 140 * rho**4 + 15 * rho**2) \
        * e.cos(2 * phi)


def quinternary_astigmatism_45(rho, phi):
    """Zernike quinternary astigmatism 45°."""
    return (210 * rho**10 - 504 * rho**8 + 420 * rho**6 - 140 * rho**4 + 15 * rho**2) \
        * e.sin(2 * phi)


def quinternary_coma_y(rho, phi):
    """Zernike quinternary coma Y."""
    return (462 * rho**11 - 1260 * rho**9 + 1260 * rho**7 - 560 * rho**5 + 105 * rho**3 - 6 * rho) \
        * e.cos(phi)


def quinternary_coma_x(rho, phi):
    """Zernike quinternary coma X."""
    return (462 * rho**11 - 1260 * rho**9 + 1260 * rho**7 - 560 * rho**5 + 105 * rho**3 - 6 * rho) \
        * e.sin(phi)


def quinternary_spherical(rho, phi):
    """Zernike quinternary spherical."""
    return 924 * rho**12 \
        - 2772 * rho**10 \
        + 3150 * rho**8 \
        - 1680 * rho**6 \
        + 420 * rho**4 \
        - 42 * rho**2 \
        + 1


def primary_septafoil_y(rho, phi):
    """Zernike primary septafoil."""
    return 4 * rho**7 * e.cos(7 * phi)


def primary_septafoil_x(rho, phi):
    """Zernike primary septafoil."""
    return 4 * rho**7 * e.sin(7 * phi)


# norms
piston.norm = 1
tip.norm = 2
tilt.norm = 2
defocus.norm = e.sqrt(3)
primary_astigmatism_00.norm = e.sqrt(6)
primary_astigmatism_45.norm = e.sqrt(6)
primary_coma_y.norm = 2 * e.sqrt(2)
primary_coma_x.norm = 2 * e.sqrt(2)
primary_spherical.norm = e.sqrt(5)
primary_trefoil_y.norm = 2 * e.sqrt(2)
primary_trefoil_x.norm = 2 * e.sqrt(2)
secondary_astigmatism_00.norm = e.sqrt(10)
secondary_astigmatism_45.norm = e.sqrt(10)
secondary_coma_y.norm = 2 * e.sqrt(3)
secondary_coma_x.norm = 2 * e.sqrt(3)
secondary_spherical.norm = e.sqrt(7)
primary_tetrafoil_y.norm = e.sqrt(10)
primary_tetrafoil_x.norm = e.sqrt(10)
secondary_trefoil_y.norm = 2 * e.sqrt(3)
secondary_trefoil_x.norm = 2 * e.sqrt(3)
tertiary_astigmatism_00.norm = e.sqrt(14)
tertiary_astigmatism_45.norm = e.sqrt(14)
tertiary_coma_y.norm = 4
tertiary_coma_x.norm = 4
tertiary_spherical.norm = 3
primary_pentafoil_y.norm = 2 * e.sqrt(3)
primary_pentafoil_x.norm = 2 * e.sqrt(3)
secondary_tetrafoil_y.norm = e.sqrt(14)
secondary_tetrafoil_x.norm = e.sqrt(14)
tertiary_trefoil_y.norm = 4
tertiary_trefoil_x.norm = 4
quaternary_astigmatism_00.norm = 3 * e.sqrt(2)
quaternary_astigmatism_45.norm = 3 * e.sqrt(2)
quaternary_coma_y.norm = 2 * e.sqrt(5)
quaternary_coma_x.norm = 2 * e.sqrt(5)
quaternary_spherical.norm = e.sqrt(11)
primary_hexafoil_y.norm = e.sqrt(14)
primary_hexafoil_x.norm = e.sqrt(14)
secondary_pentafoil_y.norm = 4
secondary_pentafoil_x.norm = 4
tertiary_tetrafoil_y.norm = 3 * e.sqrt(2)
tertiary_tetrafoil_x.norm = 3 * e.sqrt(2)
quaternary_trefoil_y.norm = 2 * e.sqrt(5)
quaternary_trefoil_x.norm = 2 * e.sqrt(5)
quinternary_astigmatism_00.norm = e.sqrt(22)
quinternary_astigmatism_45.norm = e.sqrt(22)
quinternary_coma_y.norm = 2 * e.sqrt(6)
quinternary_coma_x.norm = 2 * e.sqrt(6)
quinternary_spherical.norm = e.sqrt(13)
primary_septafoil_y.norm = e.sqrt(16)
primary_septafoil_x.norm = e.sqrt(16)

# names
piston.name = 'Piston'
tip.name = 'Tilt Y'
tilt.name = 'Tilt X'
defocus.name = 'Defocus'
primary_astigmatism_00.name = 'Primary Astigmatism 0°'
primary_astigmatism_45.name = 'Primary Astigmatism 45°'
primary_coma_y.name = 'Primary Coma Y'
primary_coma_x.name = 'Primary Coma X'
primary_spherical.name = 'Primary Spherical'
primary_trefoil_y.name = 'Primary Trefoil Y'
primary_trefoil_x.name = 'Primary Trefoil X'
secondary_astigmatism_00.name = 'Secondary Astigmatism 0°'
secondary_astigmatism_45.name = 'Secondary Astigmatism 45°'
secondary_coma_y.name = 'Secondary Coma Y'
secondary_coma_x.name = 'Secondary Coma X'
secondary_spherical.name = 'Secondary Spherical'
primary_tetrafoil_y.name = 'Primary Tetrafoil Y'
primary_tetrafoil_x.name = 'Primary Tetrafoil X'
secondary_trefoil_y.name = 'Secondary Trefoil Y'
secondary_trefoil_x.name = 'Secondary Trefoil X'
tertiary_astigmatism_00.name = 'Tertiary Astigmatism 0°'
tertiary_astigmatism_45.name = 'Tertiary Astigmatism 45°'
tertiary_coma_y.name = 'Tertiary Coma Y'
tertiary_coma_x.name = 'Tertiary Coma X'
tertiary_spherical.name = 'Tertiary Spherical'
primary_pentafoil_y.name = 'Primary Pentafoil Y'
primary_pentafoil_x.name = 'Primary Pentafoil X'
secondary_tetrafoil_y.name = 'Secondary Tetrafoil Y'
secondary_tetrafoil_x.name = 'Secondary Tetrafoil X'
tertiary_trefoil_y.name = 'Tertiary Trefoil Y'
tertiary_trefoil_x.name = 'Tertiary Trefoil X'
quaternary_astigmatism_00.name = 'Quaternary Astigmatism 0°'
quaternary_astigmatism_45.name = 'Quaternary Astigmatism 45°'
quaternary_coma_y.name = 'Quaternary Coma Y'
quaternary_coma_x.name = 'Quaternary Coma X'
quaternary_spherical.name = 'Quaternary Spherical'
primary_hexafoil_y.name = 'Primary Hexafoil Y'
primary_hexafoil_x.name = 'Primary Hexafoil X'
secondary_pentafoil_y.name = 'Secondary Pentafoil Y'
secondary_pentafoil_x.name = 'Secondary Pentafoil X'
tertiary_tetrafoil_y.name = 'Tertiary Tetrafoil Y'
tertiary_tetrafoil_x.name = 'Tertiary Tetrafoil X'
quaternary_trefoil_y.name = 'Quaternary Trefoil Y'
quaternary_trefoil_x.name = 'Quaternary Trefoil X'
quinternary_astigmatism_00.name = 'Quinternary Astigmatism 0°'
quinternary_astigmatism_45.name = 'Quinternary Astigmatism 45°'
quinternary_coma_y.name = 'Quinternary Coma Y'
quinternary_coma_x.name = 'Quinternary Coma X'
quinternary_spherical.name = 'Quinternary Spherical'
primary_septafoil_y.name = 'Primary Septafoil Y'
primary_septafoil_x.name = 'Primary Septafoil X'

zernikes = [
    piston,
    tip,
    tilt,
    defocus,
    primary_astigmatism_00,
    primary_astigmatism_45,
    primary_coma_y,
    primary_coma_x,
    primary_spherical,
    primary_trefoil_y,
    primary_trefoil_x,
    secondary_astigmatism_00,
    secondary_astigmatism_45,
    secondary_coma_y,
    secondary_coma_x,
    secondary_spherical,
    primary_tetrafoil_y,
    primary_tetrafoil_x,
    secondary_trefoil_y,
    secondary_trefoil_x,
    tertiary_astigmatism_00,
    tertiary_astigmatism_45,
    tertiary_coma_y,
    tertiary_coma_x,
    tertiary_spherical,
    primary_pentafoil_y,
    primary_pentafoil_x,
    secondary_tetrafoil_y,
    secondary_tetrafoil_x,
    tertiary_trefoil_y,
    tertiary_trefoil_x,
    quaternary_astigmatism_00,
    quaternary_astigmatism_45,
    quaternary_coma_y,
    quaternary_coma_x,
    quaternary_spherical,
    primary_hexafoil_y,
    primary_hexafoil_x,
    secondary_pentafoil_y,
    secondary_pentafoil_x,
    tertiary_tetrafoil_y,
    tertiary_tetrafoil_x,
    quaternary_trefoil_y,
    quaternary_trefoil_x,
    quinternary_astigmatism_00,
    quinternary_astigmatism_45,
    quinternary_coma_y,
    quinternary_coma_x,
    quinternary_spherical,
    primary_septafoil_y,
    primary_septafoil_x,
]

zernikes_cpu = [jit(zernikes[0])]
for func in zernikes[1:]:
    compfunc = vectorize(func)
    compfunc.name = func.name
    compfunc.norm = func.norm
    zernikes_cpu.append(compfunc)

zernikes_gpu = [fuse(func) for func in zernikes[1:]]  # cupy compiled zernikes
zernikes_gpu.insert(0, zernikes[0])


def change_backend(to):
    """Change the backend between cuda/cupy and CPU."""
    if to == 'cu':
        globals()['zernikes'] = zernikes_gpu
    elif to == 'np':
        globals()['zernikes'] = zernikes_cpu


config.chbackend_observers.append(change_backend)


fringemap = {
    0: 0,
    1: 1,
    2: 2,
    3: 3,
    4: 4,
    5: 5,
    6: 6,
    7: 7,
    8: 8,
    9: 9,
    10: 10,
    11: 11,
    12: 12,
    13: 13,
    14: 14,
    15: 15,
    16: 16,
    17: 17,
    18: 18,
    19: 19,
    20: 20,
    21: 21,
    22: 22,
    23: 23,
    24: 24,
    25: 25,
    26: 26,
    27: 27,
    28: 28,
    29: 29,
    30: 30,
    31: 31,
    32: 32,
    33: 33,
    34: 34,
    35: 35,
    36: 36,
    37: 37,
    38: 38,
    39: 39,
    40: 40,
    41: 41,
    42: 42,
    43: 43,
    44: 44,
    45: 45,
    46: 46,
    47: 47,
    48: 48,
}
nollmap = {
    0: 0,
    1: 1,
    2: 2,
    3: 3,
    4: 4,
    5: 5,
    6: 6,
    7: 7,
    8: 9,
    9: 10,
    10: 8,
    11: 11,
    12: 12,
    13: 16,
    14: 17,
    15: 13,
    16: 14,
    17: 18,
    18: 19,
    19: 25,
    20: 26,
    21: 15,
    22: 20,
    23: 21,
    24: 27,
    25: 28,
    26: 36,
    27: 37,
    28: 22,
    29: 23,
    30: 29,
    31: 30,
    32: 38,
    33: 39,
    34: 49,
    35: 50,
    36: 24,
}
maps = {
    'fringe': fringemap,
    'noll': nollmap,
}


def zernikename(idx, base, map_):
    """Return the name of a Fringe Zernike with the given index and base."""
    return zernikes[map_[idx-base]].name


def zernikes_to_magnitude_angle(coefs, namer):
    """Convert Fringe Zernike polynomial set to a magnitude and phase representation."""
    def mkary():  # default for defaultdict
        return e.zeros(2)

    # make a list of names to go with the coefficients
    names = [namer(i, base=0) for i in range(len(coefs))]
    combinations = defaultdict(mkary)

    # for each name and coefficient, make a len 2 array.  Put the Y or 0 degree values in the first slot
    for coef, name in zip(coefs, names):
        if name.endswith(('X', 'Y', '°')):
            newname = ' '.join(name.split(' ')[:-1])
            if name.endswith('Y'):
                combinations[newname][0] = coef
            elif name.endswith('X'):
                combinations[newname][1] = coef
            elif name[-2] == '5':  # 45 degree case
                combinations[newname][1] = coef
            else:
                combinations[newname][0] = coef
        else:
            combinations[name][0] = coef

    # now go over the combinations and compute the L2 norms and angles
    for name in combinations:
        ovals = combinations[name]
        magnitude = e.sqrt((ovals**2).sum())
        if 'Spheric' in name or 'focus' in name or 'iston' in name:
            phase = 0
        else:
            phase = e.degrees(e.arctan2(*ovals))
        values = (magnitude, phase)
        combinations[name] = values

    return dict(combinations)  # cast to regular dict for return


zernikefuncs = {
    'name': {
        'fringe': partial(zernikename, map_=fringemap),
        'noll': partial(zernikename, map_=nollmap),
    }
}
zernikefuncs.update({
    'magnitude_angle': {
        'fringe': partial(zernikes_to_magnitude_angle, namer=zernikefuncs['name']['fringe']),
        'noll': partial(zernikes_to_magnitude_angle, namer=zernikefuncs['name']['noll']),
    }
})


def zernike_norm(n, m):
    """Norm of a Zernike polynomial with n, m indexing."""
    return e.sqrt((2 * (n + 1)) / (1 + kronecker(m, 0)))


def n_m_to_fringe(n, m):
    """Convert (n,m) two term index to Fringe index."""
    term1 = (1 + (n + abs(m))/2)**2
    term2 = 2 * abs(m)
    term3 = (1 + sign(m)) / 2
    return int(term1 - term2 + term3)


def n_m_to_ansi_j(n, m):
    """Convert (n,m) two term index to ANSI single term index."""
    return int((n * (n + 2) + m) / 2)


def ansi_j_to_n_m(idx):
    """Convert ANSI single term to (n,m) two-term index."""
    n = int(e.ceil((-3 + e.sqrt(9 + 8*idx))/2))
    m = 2 * idx - n * (n + 2)
    return n, m


def noll_to_n_m(idx):
    """Convert Noll Z to (n, m) two-term index."""
    # I don't really understand this code, the math is inspired by POPPY
    # azimuthal order
    n = int(e.ceil((-1 + e.sqrt(1 + 8 * idx)) / 2) - 1)
    if n == 0:
        m = 0
    else:
        # this is sort of a rising factorial to use that term incorrectly
        nseries = (n + 1) * (n + 2) / 2

        res = idx - nseries - 1

        if is_odd(idx):
            sign = -1
        else:
            sign = 1

        if is_odd(n):
            ms = [1, 1]
        else:
            ms = [0]

        for i in range(n // 2):
            ms.append(ms[-1] + 2)
            ms.append(ms[-1])

        m = ms[res] * sign

    return n, m


def zero_separation(n):
    """Zero separation in normalized r based on radial order n."""
    return 1 / n ** 2


class ZCache(object):
    """Cache of Zernike terms evaluated over the unit circle."""
    def __init__(self):
        """Create a new ZCache instance."""
        self.normed = defaultdict(dict)
        self.regular = defaultdict(dict)

    def get_zernike(self, number, norm, samples):
        """Get an array of phase values for a given index, norm, and number of samples."""
        if norm is True:
            target = self.normed
        else:
            target = self.regular

        try:
            zern = target[samples][number]
        except KeyError:
            rho, phi = make_rho_phi_grid(samples, aligned='y')
            func = zernikes[number]
            zern = func(rho, phi)
            if norm is True:
                zern *= func.norm

            target[samples][number] = zern.copy()

        return zern

    def __call__(self, number, norm, samples):
        """Get an array of phase values for a given index, norm, and number of samples."""
        return self.get_zernike(number, norm, samples)

    def clear(self, *args):
        """Empty the cache."""
        self.normed = defaultdict(dict)
        self.regular = defaultdict(dict)


zcache = ZCache()
config.chbackend_observers.append(zcache.clear)


class ZCacheMN:
    """Cache of Zernike terms evaluated over the unit circle, based on (n, m) indices."""
    def __init__(self, gridcache=gridcache):
        """Create a new ZCache instance."""
        self.normed = {}
        self.regular = {}
        self.jac = {}
        self.cos = {}
        self.sin = {}
        self.gridcache = gridcache

    @retry(tries=2)
    def get_zernike(self, n, m, samples, norm):
        """Get an array of phase values for a given radial order n, azimuthal order m, number of samples, and orthonormalization."""  # NOQA
        if is_odd(n - m):
            raise ValueError('Zernike polynomials are only defined for n-m even.')
        key = (n, m, samples)
        if norm:
            d_ = self.normed
        else:
            d_ = self.regular

        try:
            return d_[key]
        except KeyError as e:
            zern = self.get_term(n=n, m=m, samples=samples)
            if norm:
                zern = zern * zernike_norm(n=n, m=m)

            d_[key] = zern
            raise e

    def get_term(self, n, m, samples):
        am = abs(m)
        r, p = self.get_grid(samples=samples, modified=False)
        term = self.get_jacobi(n=n, m=am, samples=samples)

        if m != 0:
            azterm = self.get_azterm(m=m, samples=samples)
            rterm = r ** am
            term = term * azterm * rterm

        return term

    def __call__(self, n, m, samples, norm):
        return self.get_zernike(n=n, m=m, samples=samples, norm=norm)

    @retry(tries=2)
    def get_azterm(self, m, samples):
        key = (m, samples)
        if sign(m) == -1:
            d_ = self.sin
            func = e.sin
        else:
            d_ = self.cos
            func = e.cos

        try:
            return d_[key]
        except KeyError as err:
            _, p = self.get_grid(samples=samples, modified=False)
            d_[key] = func(m * p)
            raise err

    @retry(tries=3)
    def get_jacobi(self, n, m, samples, nj=None):
        if nj is None:
            nj = (n - m) // 2
        key = (nj, m, samples)
        try:
            return self.jac[key]
        except KeyError as e:
            r, _ = self.get_grid(samples=samples)
            if nj > 2:
                jnm1 = self.get_jacobi(n=None, nj=nj - 1, m=m, samples=samples)
                jnm2 = self.get_jacobi(n=None, nj=nj - 2, m=m, samples=samples)
            else:
                jnm1, jnm2 = None, None
            jac = jacobi(nj, alpha=0, beta=m, Pnm1=jnm1, Pnm2=jnm2, x=r)
            self.jac[key] = jac
            raise e

    def get_grid(self, samples, modified=True):
        if modified:
            res = self.gridcache(samples=samples, radius=1, r='r -> 2r^2 - 1', t='t')
        else:
            res = self.gridcache(samples=samples, radius=1, r='r', t='t')

        return res['r'], res['t']

    def clear(self, *args):
        """Empty the cache."""
        self.normed = {}
        self.regular = {}
        self.jac = {}
        self.sin = {}
        self.cos = {}

    def nbytes(self):
        """total size in memory of the cache in bytes."""
        total = 0
        for dict_ in (self.normed, self.regular, self.jac, self.sin, self.cos):
            for key in dict_:
                total += dict_[key].nbytes

        return total


zcachemn = ZCacheMN()
config.chbackend_observers.append(zcachemn.clear)


class BaseZernike(Pupil):
    """Basic class implementing Zernike features."""
    _map = None
    _namer = None
    _magnituder = None
    _cache = None
    _name = None

    def __init__(self, *args, **kwargs):
        """Initialize a new Zernike instance."""
        if args is not None:
            if len(args) == 0:
                self.coefs = e.zeros(len(self._map), dtype=config.precision)
            else:
                self.coefs = e.asarray([*args[0]], dtype=config.precision)

        self.normalize = False
        pass_args = {}

        self.base = config.zernike_base
        try:
            bb = kwargs['base']
            if bb > 1:
                raise ValueError('It violates convention to use a base greater than 1.')
            elif bb < 0:
                raise ValueError('It is nonsensical to use a negative base.')
            self.base = bb
        except KeyError:
            # user did not specify base
            pass

        if kwargs is not None:
            for key, value in kwargs.items():
                if key[0].lower() == 'z' and key[1].isnumeric():
                    idx = int(key[1:])  # strip 'Z' from index
                    self.coefs[idx - self.base] = value
                elif key.lower() == 'norm':
                    self.normalize = value
                elif key.lower() == 'base':
                    self.base = value
                else:
                    pass_args[key] = value

        super().__init__(**pass_args)

    def build(self):
        """Use the wavefront coefficients stored in this class instance to build a wavefront model.

        Returns
        -------
        self : `BaseZernike`
            this Zernike instance

        """
        # build a coordinate system over which to evaluate this function
        self.phase = e.zeros((self.samples, self.samples), dtype=config.precision)
        for term, coef in enumerate(self.coefs):
            # short circuit for speed
            if coef == 0:
                continue
            else:
                idx = self._map[term]
                self.phase += coef * self._cache(idx, self.normalize, self.samples)  # NOQA

        return self

    def top_n(self, n=5):
        """Identify the top n terms in the wavefront.

        Parameters
        ----------
        n : `int`, optional
            identify the top n terms.

        Returns
        -------
        `list`
            list of tuples (magnitude, index, term)

        """
        coefs = e.asarray(self.coefs)
        coefs_work = abs(coefs)
        oidxs = e.arange(len(coefs), dtype=int) + self.base  # "original indexes"
        idxs = e.argpartition(coefs_work, -n)[-n:]  # argpartition does some magic to identify the top n (unsorted)
        idxs = idxs[e.argsort(coefs_work[idxs])[::-1]]  # use argsort to sort them in ascending order and reverse
        big_terms = coefs[idxs]  # finally, take the values from the
        big_idxs = oidxs[idxs]
        names = e.asarray(self.names, dtype=str)[big_idxs - self.base]
        return list(zip(big_terms, big_idxs, names))

    @property
    def magnitudes(self):
        """Return the magnitude and angles of the zernike components in this wavefront."""
        # need to call through class variable to avoid insertion of self as arg
        return self.__class__._magnituder(self.coefs)  # NOQA

    @property
    def names(self):
        """Names of the terms in self."""
        # need to call through class variable to avoid insertion of self as arg
        idxs = e.asarray(range(len(self.coefs))) + self.base
        return [self.__class__._namer(i, base=self.base) for i in idxs]  # NOQA

    def barplot(self, orientation='h', buffer=1, zorder=3, number=True, offset=0, width=0.8, fig=None, ax=None):
        """Create a barplot of coefficients and their names.

        Parameters
        ----------
        orientation : `str`, {'h', 'v', 'horizontal', 'vertical'}
            orientation of the plot
        buffer : `float`, optional
            buffer to use around the left and right (or top and bottom) bars
        zorder : `int`, optional
            zorder of the bars.  Use zorder > 3 to put bars in front of gridlines
        number : `bool`, optional
            if True, plot numbers along the y=0 line showing indices
        offset : `float`, optional
            offset to apply to bars, useful for before/after Zernike breakdowns
        width : `float`, optional
            width of bars, useful for before/after Zernike breakdowns
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        """
        from matplotlib import pyplot as plt
        fig, ax = share_fig_ax(fig, ax)

        coefs = e.asarray(self.coefs)
        idxs = e.asarray(range(len(coefs))) + self.base
        names = self.names
        lab = self.labels.z(self.xy_unit, self.z_unit)
        lims = (idxs[0] - buffer, idxs[-1] + buffer)
        if orientation.lower() in ('h', 'horizontal'):
            vmin, vmax = coefs.min(), coefs.max()
            drange = vmax - vmin
            offsetY = drange * 0.01

            ax.bar(idxs + offset, self.coefs, zorder=zorder, width=width)
            plt.xticks(idxs, names, rotation=90)
            if number:
                for i in idxs:
                    ax.text(i, offsetY, str(i), ha='center')
            ax.set(ylabel=lab, xlim=lims)
        else:
            ax.barh(idxs + offset, self.coefs, zorder=zorder, height=width)
            plt.yticks(idxs, names)
            if number:
                for i in idxs:
                    ax.text(0, i, str(i), ha='center')
            ax.set(xlabel=lab, ylim=lims)
        return fig, ax

    def barplot_magnitudes(self, orientation='h', sort=False,
                           buffer=1, zorder=3, offset=0, width=0.8,
                           fig=None, ax=None):
        """Create a barplot of magnitudes of coefficient pairs and their names.

        E.g., astigmatism will get one bar.

        Parameters
        ----------
        orientation : `str`, {'h', 'v', 'horizontal', 'vertical'}
            orientation of the plot
        sort : `bool`, optional
            whether to sort the zernikes in descending order
        buffer : `float`, optional
            buffer to use around the left and right (or top and bottom) bars
        zorder : `int`, optional
            zorder of the bars.  Use zorder > 3 to put bars in front of gridlines
        offset : `float`, optional
            offset to apply to bars, useful for before/after Zernike breakdowns
        width : `float`, optional
            width of bars, useful for before/after Zernike breakdowns
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        """
        from matplotlib import pyplot as plt

        magang = self.magnitudes
        mags = [m[0] for m in magang.values()]
        names = magang.keys()
        idxs = e.asarray(list(range(len(names))))

        if sort:
            mags, names = sort_xy(mags, names)
            mags = list(reversed(mags))
            names = list(reversed(names))
        lab = self.labels.z(self.xy_unit, self.z_unit)
        lims = (idxs[0] - buffer, idxs[-1] + buffer)
        fig, ax = share_fig_ax(fig, ax)
        if orientation.lower() in ('h', 'horizontal'):
            ax.bar(idxs + offset, mags, zorder=zorder, width=width)
            plt.xticks(idxs, names, rotation=90)
            ax.set(ylabel=lab, xlim=lims)
        else:
            ax.barh(idxs + offset, mags, zorder=zorder, height=width)
            plt.yticks(idxs, names)
            ax.set(xlabel=lab, ylim=lims)
        return fig, ax

    def barplot_topn(self, n=5, orientation='h', buffer=1, zorder=3, fig=None, ax=None):
        """Plot the top n terms in the wavefront.

        Parameters
        ----------
        n : `int`, optional
            plot the top n terms.
        orientation : `str`, {'h', 'v', 'horizontal', 'vertical'}
            orientation of the plot
        buffer : `float`, optional
            buffer to use around the left and right (or top and bottom) bars
        zorder : `int`, optional
            zorder of the bars.  Use zorder > 3 to put bars in front of gridlines
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        Returns
        -------
        fig : `matplotlib.figure.Figure`
            Figure containing the plot
        ax : `matplotlib.axes.Axis`
            Axis containing the plot

        """
        from matplotlib import pyplot as plt

        topn = self.top_n(n)
        magnitudes = [n[0] for n in topn]
        names = [n[2] for n in topn]
        idxs = range(len(names))

        fig, ax = share_fig_ax(fig, ax)

        lab = self.labels.z(self.xy_unit, self.z_unit)
        lims = (idxs[0] - buffer, idxs[-1] + buffer)
        if orientation.lower() in ('h', 'horizontal'):
            ax.bar(idxs, magnitudes, zorder=zorder)
            plt.xticks(idxs, names, rotation=90)
            ax.set(ylabel=lab, xlim=lims)
        else:
            ax.barh(idxs, magnitudes, zorder=zorder)
            plt.yticks(idxs, names)
            ax.set(xlabel=lab, ylim=lims)
        return fig, ax

    def truncate(self, n):
        """Truncate the wavefront to the first n terms.

        Parameters
        ----------
        n : `int`
            number of terms to keep.

        Returns
        -------
        `self`
            modified FringeZernike instance.

        """
        if n > len(self.coefs):
            return self
        else:
            self.coefs = self.coefs[:n]
            self.build()
            return self

    def truncate_topn(self, n):
        """Truncate the pupil to only the top n terms.

        Parameters
        ----------
        n : `int`
            number of parameters to keep

        Returns
        -------
        `self`
            modified FringeZernike instance.

        """
        topn = self.top_n(n)
        new_coefs = e.zeros(len(self.coefs), dtype=config.precision)
        for coef in topn:
            mag, index, *_ = coef
            new_coefs[index-self.base] = mag

        self.coefs = new_coefs
        self.build()
        return self

    def __str__(self):
        """Pretty-print pupil description."""
        if self.normalize is True:
            header = f'rms normalized {self._name} Zernike description with:\n\t'
        else:
            header = f'{self._name} Zernike description with:\n\t'

        strs = []
        for number, coef in enumerate(self.coefs):
            # skip 0 terms
            if coef == 0:
                continue

            # positive coefficient, prepend with +
            if e.sign(coef) == 1:
                _ = '+' + f'{coef:.3f}'
            # negative, sign comes from the value
            else:
                _ = f'{coef:.3f}'

            # create the name
            idx = self._map[number]
            name = f'Z{number+self.base} - {zernikes[idx].name}'

            strs.append(' '.join([_, name]))
        body = '\n\t'.join(strs)
        unit_str = f'{self.labels.unit_prefix}{self.z_unit}{self.labels.unit_suffix}'
        footer = f'\n\t{self.pv:.3f} PV, {self.rms:.3f} RMS {unit_str}'
        return f'{header}{body}{footer}'


class FringeZernike(BaseZernike):
    """Fringe Zernike description of an optical pupil."""
    _map = fringemap
    _cache = zcache
    _magnituder = zernikefuncs['magnitude_angle']['fringe']
    _namer = zernikefuncs['name']['fringe']
    _name = 'Fringe'


class NollZernike(BaseZernike):
    """Noll Zernike description of an optical pupil."""
    _map = nollmap
    _cache = zcache
    _magnituder = zernikefuncs['magnitude_angle']['noll']
    _namer = zernikefuncs['name']['noll']
    _name = 'Noll'


class ANSI2TermZernike(Pupil):
    """2-term ANSI Zernike description of an optical pupil."""
    _cache = zcachemn

    def __init__(self, *args, **kwargs):
        """Initialize a new Zernike instance."""
        self.normalize = True
        pass_args = {}

        self.terms = []
        if kwargs is not None:
            for key, value in kwargs.items():
                k0l = key[0].lower()
                if k0l in ('a', 'b'):
                    # the only kwarg to contain a _ is a Zernike term
                    # the kwarg looks like A<n>_<m>=<coef>

                    # if the term is "A", it is a cosine and m is positive
                    if k0l == 'a':
                        msign = 1
                    elif k0l == 'b':
                        msign = -1

                    if '_' in key:
                        front, back = key.split('_')
                        n = int(front[1:])
                        m = int(back) * msign
                    else:
                        n = int(key[1:])
                        m = 0

                    self.terms.append((n, m, value))  # coef = value

                elif key.lower() == 'norm':
                    self.normalize = value
                else:
                    pass_args[key] = value

        super().__init__(**pass_args)

    def build(self):
        """Use the wavefront coefficients stored in this class instance to build a wavefront model.

        Returns
        -------
        self : `BaseZernike`
            this Zernike instance

        """
        # build a coordinate system over which to evaluate this function
        self.phase = e.zeros((self.samples, self.samples), dtype=config.precision)
        for (n, m, coef) in self.terms:
            # short circuit for speed
            if coef == 0:
                continue
            else:
                zernike = self._cache(n=n, m=m, samples=self.samples, norm=self.normalize) * coef
                self.phase += zernike

        return self


class ANSI1TermZernike(Pupil):
    """1-term ANSI Zernike description of an optical pupil."""
    _cache = zcachemn

    def __init__(self, *args, **kwargs):
        """Initialize a new Zernike instance."""
        self.normalize = True
        pass_args = {}

        self.terms = {}
        if kwargs is not None:
            for key, value in kwargs.items():
                if key[0].lower() == 'z':
                    self.terms[int(key[1:])] = value

                elif key.lower() == 'norm':
                    self.normalize = value
                else:
                    pass_args[key] = value

        super().__init__(**pass_args)

    def build(self):
        """Use the wavefront coefficients stored in this class instance to build a wavefront model.

        Returns
        -------
        self : `BaseZernike`
            this Zernike instance

        """
        # build a coordinate system over which to evaluate this function
        self.phase = e.zeros((self.samples, self.samples), dtype=config.precision)
        for idx, coef in self.terms.items():
            # short circuit for speed
            if coef == 0:
                continue
            else:
                n, m = ansi_j_to_n_m(idx)
                zernike = self._cache(n=n, m=m, samples=self.samples, norm=self.normalize) * coef
                self.phase += zernike

        return self


def zernikefit(data, x=None, y=None,
               rho=None, phi=None, terms=16,
               norm=False, residual=False,
               round_at=6, map_='fringe'):
    """Fits a number of Zernike coefficients to provided data.

    Works by minimizing the mean square error  between each coefficient and the
    given data.  The data should be uniformly sampled in an x,y grid.

    Parameters
    ----------
    data : `numpy.ndarray`
        data to fit to.
    x : `numpy.ndarray`, optional
        x coordinates, same shape as data
    y : `numpy.ndarray`, optional
        y coordinates, same shape as data
    rho : `numpy.ndarray`, optional
        radial coordinates, same shape as data
    phi : `numpy.ndarray`, optional
        azimuthal coordinates, same shape as data
    terms : `int`, optional
        number of terms to fit, fits terms 0~terms
    norm : `bool`, optional
        if True, normalize coefficients to unit RMS value
    residual : `bool`, optional
        if True, return a tuple of (coefficients, residual)
    round_at : `int`, optional
        decimal place to round values at.
    map_ : `str`, optional, {'fringe', 'noll'}
        which ordering of Zernikes to use

    Returns
    -------
    coefficients : `numpy.ndarray`
        an array of coefficients matching the input data.
    residual : `float`
        RMS error between the input data and the fit.

    Raises
    ------
    ValueError
        too many terms requested.

    """
    map_ = maps[map_]
    if terms > len(fringemap):
        raise ValueError(f'number of terms must be less than {len(fringemap)}')

    data = data.T  # transpose to mimic transpose of zernikes

    # precompute the valid indexes in the original data
    pts = e.isfinite(data)

    if x is None and rho is None:
        # set up an x/y rho/phi grid to evaluate Zernikes on
        rho, phi = make_rho_phi_grid(*reversed(data.shape))
        rho = rho[pts].flatten()
        phi = phi[pts].flatten()
    elif rho is None:
        rho, phi = cart_to_polar(x, y)
        rho, phi = rho[pts].flatten(), phi[pts].flatten()

    # compute each Zernike term
    zerns_raw = []
    for i in range(terms):
        func = zernikes[map_[i]]
        base_zern = func(rho, phi)
        if norm:
            base_zern *= func.norm
        zerns_raw.append(base_zern)
    zerns = e.asarray(zerns_raw).T

    # use least squares to compute the coefficients
    meas_pts = data[pts].flatten()
    coefs = e.linalg.lstsq(zerns, meas_pts, rcond=None)[0]
    if round_at is not None:
        coefs = coefs.round(round_at)

    if residual is True:
        components = []
        for zern, coef in zip(zerns_raw, coefs):
            components.append(coef * zern)

        _fit = e.asarray(components)
        _fit = _fit.sum(axis=0)
        rmserr = rms(data[pts].flatten() - _fit)
        return coefs, rmserr
    else:
        return coefs
